/*
 * eeprom.cc -- USB Atmel programmer FTDI EEPROM editing routines. 
 * 
 * Copyright (c) 2006--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <ftdi.h>
#include <ctype.h>

#include "usbio.h"
#include "eeprom.h"


static void DumpEEPROM(const struct ftdi_eeprom *e)
{
	printf("FTDI configuration EEPROM content:\n"
		"  vendor:          0x%04x\n"
		"  product:         0x%04x\n"
		"  self_powered:    %d\n"
		"  remote_wakeup:   %d\n"
		"  BM_type_chip:    %d\n"
		"  isochronous:     in: %d,  out: %d\n"
		"  susp_pull_downs: %d\n"
		"  use_serial:      %d\n"
		"  usb_version:     %d,  change_usb_version: %d\n"
		"  max_power:       %d\n"
		"  manufacturer:    %s%s%s\n"
		"  product:         %s%s%s\n"
		"  serial:          %s%s%s\n"
		,e->vendor_id,e->product_id,
		e->self_powered,e->remote_wakeup,e->BM_type_chip,
		e->in_is_isochronous,e->out_is_isochronous,e->suspend_pull_downs,
		e->use_serial,e->usb_version,e->change_usb_version,e->max_power,
		e->manufacturer ? "\"" : "",
			e->manufacturer ? e->manufacturer  : "[none]",
			e->manufacturer ? "\"" : "",
		e->product ? "\"" : "",
			e->product ? e->product : "[none]",
			e->product ? "\"" : "",
		e->serial ? "\"" : "",
			e->serial ? e->serial : "[none]",
			e->serial ? "\"" : "");
}


static void SetEEPROMString(struct ftdi_eeprom *e,const char *str,char which)
{
	char **dest=NULL;
	switch(which)
	{
		case 'm':  dest=&e->manufacturer;  break;
		case 's':  dest=&e->serial;        break;
		case 'p':  dest=&e->product;       break;
		default:  assert(0);
	}
	
	if(*dest)
	{  free(*dest);  *dest=NULL;  }
	if(str)
	{
		*dest=(char*)malloc(strlen(str)+1);
		strcpy(*dest,str);
	}
}


static void SetEEPROMDefaults(struct ftdi_eeprom *e)
{
	SetEEPROMString(e,NULL,'m');
	SetEEPROMString(e,NULL,'p');
	SetEEPROMString(e,NULL,'s');
	ftdi_eeprom_initdefaults(e);
}


static void SetEEPROMForProgrammer(struct ftdi_eeprom *e)
{
	SetEEPROMDefaults(e);
	
	e->vendor_id=0x0403;
	e->product_id=0x6002;
	e->self_powered=0;
	//e->remote_wakeup=1;
	//e->BM_type_chip=1;
	e->in_is_isochronous=0;
	e->out_is_isochronous=0;
	e->suspend_pull_downs=0;
	e->use_serial=0;
	//e->usb_version=0;
	e->change_usb_version=0;
	e->max_power=250;
	//SetEEPROMString(e,"WieserLabs (01/2006)",'m');
	SetEEPROMString(e,"USB-AtmelPrg",'p');
	//SetEEPROMString(e,"Rev. 2/2",'s');
}


int USB_EditFTDIEEPROM(void)
{
	printf("Reading current EEPROM content...\n");
	struct ftdi_eeprom *eep=USB_ReadFTDIEEPROM();
	
	DumpEEPROM(eep);
	printf("FTDI EEPROM Editor: enter command; `?' for help, ^D to quit.\n");
	printf("[development version; not yet very conventient]\n");
	printf(
		"\n"
		"*** NOTE: Will work on FT2xxB chips but                     ***\n"
		"***       will _not_ function for the newer FT2xxR variant! ***\n"
		"\n");
	
	for(;;)
	{
		char cmd_buf[256];
		printf("eeprom> ");  fflush(stdout);
		if(!fgets(cmd_buf,256,stdin))
		{
			if(feof(stdin))
			{  printf("^D\n");  return(0);  }
			fprintf(stderr,"reading stdin: %s\n",strerror(errno));
			return(1);
		}
		char *cmd=cmd_buf;
		while(isspace(*cmd))  ++cmd;
		if(*cmd=='\0')  continue;
		char *end=cmd+strlen(cmd);
		while(isspace(end[-1]))  --end;
		*end='\0';
		
		if(!strcmp(cmd,"help") || !strcmp(cmd,"?"))
		{
			printf("Available commands for EEPROM editing:\n"
				"  ?          print this help\n"
				"  print      print current (local copy of the) EEPROM content\n"
				"  quit/^D    quit EEPROM editor without saving\n"
				"  read       re-read current EEPROM content\n"
				"  write      actually write EEPROM content\n"
				"  erase      just erase EEPROM (implied by write)\n"
				"  defaults   set default EEPROM content\n"
				"  atmelprg   set defaults for usb-atmelprg\n"
				"  vendorid   change vendor code\n"
				"  productid  change product code\n"
				"  manuf      change manufacturer string\n"
				"  serial     change serial string\n"
				"  product    change product string\n"
				"  maxpwr     change maximum power value (0..250)\n"
				"Note that no changes are committed until \"write\".\n");
		}
		else if(!strcmp(cmd,"print"))
		{  DumpEEPROM(eep);  }
		else if(!strcmp(cmd,"quit"))
		{  return(0);  }
		else if(!strcmp(cmd,"read"))
		{
			USB_FreeFTDIEEPROM(eep);
			printf("Reading current EEPROM content...\n");
			eep=USB_ReadFTDIEEPROM();
		}
		else if(!strcmp(cmd,"write"))
		{
			if(!USB_WriteFTDIEEPROM(eep))
			{  printf("Current config written to EEPROM on "
				"programmer board.\n");  }
		}
		else if(!strcmp(cmd,"erase"))
		{
			if(!USB_WriteFTDIEEPROM(NULL))
			{  printf("Config EEPROM on programmer board erased.\n");  }
		}
		else if(!strcmp(cmd,"vendorid") || !strcmp(cmd,"productid"))
		{
			char which=cmd[0];
			
			printf("Enter new %s code> ",
				which=='v' ? "vendor" : "product");
			fflush(stdout);
			
			*cmd_buf='\0';
			if(fgets(cmd_buf,256,stdin))
			{
				int code=strtol(cmd_buf,NULL,0);
				
				if(which=='v')
				{  eep->vendor_id=code;  }
				else
				{  eep->product_id=code;  }
			}
			else
			{  printf("\n");  }
		}
		else if(!strcmp(cmd,"maxpwr"))
		{
			printf("Enter new max_power value (0..250)> ");
			fflush(stdout);
			
			*cmd_buf='\0';
			if(fgets(cmd_buf,256,stdin))
			{
				int val=strtol(cmd_buf,NULL,0);
				
				eep->max_power=val;
			}
			else
			{  printf("\n");  }
		}
		else if(!strcmp(cmd,"manuf") || 
		        !strcmp(cmd,"serial") || 
		        !strcmp(cmd,"product"))
		{
			char which=cmd[0];
			
			printf("Enter new %s string> ",
				which=='m' ? "manufacturer" : "serial");
			fflush(stdout);
			
			*cmd_buf='\0';
			if(fgets(cmd_buf,256,stdin))
			{
				char *end=cmd+strlen(cmd);
				while(end>cmd && isspace(end[-1]))  --end;
				*end='\0';
				
				SetEEPROMString(eep,cmd_buf,which);
			}
			else
			{  printf("\n");  }
		}
		else if(!strcmp(cmd,"defaults"))
		{
			SetEEPROMDefaults(eep);
			DumpEEPROM(eep);
		}
		else if(!strcmp(cmd,"atmelprg"))
		{
			SetEEPROMForProgrammer(eep);
			DumpEEPROM(eep);
		}
		else
		{
			fprintf(stderr,"Invalid command \"%s\". Try `?' for help.\n",
				cmd);
		}
	}
	
	USB_FreeFTDIEEPROM(eep);
	
	return(0);
}
