/*
 * usbio.c -- USB Atmel programmer USB access routines. 
 * 
 * Copyright (c) 2005--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

// NOTE: This is a C and NOT a C++ file. 
// This is required because in C, sizeof(struct{})=0 while in C++ it is 1. 
// (At least with GNU C, even if attribute __packed__ is applied.)

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "usbio.h"

//------------------------------------------------------------------------------
#define USB_ON_HOST_SIDE 1
#include "usb_types.h"
//------------------------------------------------------------------------------

// Global context. 
USBContext *usb_context=NULL;

struct _USBContext
{
	struct ftdi_context ctx;
	int ctx_valid;  // 2 -> must close; 1 -> must deinit; 0 -> invalid. 
	
	AdapterType atype;
	int enable_nibble_io;
	
	int reading_hdr_bytes;  // -1 -> body
	int body_read_done;
	
	char usb_idf_str[256];  // "0403:6002 (USB-Atmelprg, Rev. XX)"
	
	unsigned char read_nibble_carry;  // Highest bit is use flag. 
	
	char usb_recv_buf[256];  // Packet cannot be longer (1 byte size field). 
	struct USBRecvPacket *recv_pack;  // Pointer to usb_recv_buf. 
};


const char *USB_Status_String(USB_Status status)
{
	switch(status)
	{
		case USB_OK:            return("success");
		case USB_Timeout:       return("timeout");
		case USB_Interrupted:   return("interrupted");
		case USB_Unexpected:    return("unexpected reply");
		case USB_ErrorReply:    return("error reply");
		case USB_UnknownBoard:  return("unknown board");
		case USB_ConnectFailed: return("failed to connect");
		case USB_WriteFailed:   return("Atmel write failure");
		default:  return("???");
	}
	
	return("???");
}


void Error(const char *where,int rv,int doclose,int doexit)
{
	if(rv)
	{
		fprintf(stderr,"%s: rv=%d, error=%s, errno=%s (%d)\n",
			where,rv,usb_context->ctx.error_str,strerror(errno),errno);
		if(doclose && usb_context->ctx_valid>=2)
		{
			rv=ftdi_usb_close(&usb_context->ctx);
			usb_context->ctx_valid=1;
			fprintf(stderr,"close: rv=%d, error=%s, errno=%s (%d)\n",
				rv,usb_context->ctx.error_str,strerror(errno),errno);
		}
		if(usb_context->ctx_valid>=1)
		{
			ftdi_deinit(&usb_context->ctx);
			usb_context->ctx_valid=0;
		}
		if(doexit)
		{  free(usb_context);  usb_context=NULL;  exit(1);  }
	}
	else if(verbose>=2)
	{  fprintf(stderr,"%s: OK\n",where);  }
}


static int USB_DoWriteData(unsigned char *buf,int size)
{
	if(usb_context->enable_nibble_io)
	{
		// Convert into nibble-bytes first. 
		unsigned char buf2[size*2];
		int i;
		for(i=0; i<size; i++)
		{
			unsigned char tmp=buf[i];
			buf2[2*i]=tmp<<4;
			buf2[2*i+1]=tmp&0xf0U;
		}
		
		if(verbose>3)
		{  fprintf(stderr,"ftdi_write_data(%u nibble-bytes)= ",2*size);  }
		int rv=ftdi_write_data(&usb_context->ctx,buf2,2*size);
		if(verbose>3)
		{  fprintf(stderr,"%d\n",rv);  }
		if(rv<2*size)  Error("write",rv,1,1);
		
		//if(verbose>2)
		//{  fprintf(stderr,"wrote %d bytes\n",2*size);  }
		
		return(rv!=2*size);
	}
	else
	{
		if(verbose>3)
		{  fprintf(stderr,"ftdi_write_data(%u bytes)= ",size);  }
		int rv=ftdi_write_data(&usb_context->ctx,buf,size);
		if(verbose>3)
		{  fprintf(stderr,"%d\n",rv);  }
		if(rv<size)  Error("write",rv,1,1);
		
		return(rv!=size);
	}
}


// Returns -1 on error. 
static int USB_DoReadData(unsigned char *buf,int size)
{
	assert(size>=0);
	
	if(usb_context->enable_nibble_io)
	{
		unsigned char buf2[2*size+1];
		
		if(verbose>3)
		{  fprintf(stderr,"ftdi_read_data(%u nibble-bytes)= ",2*size);  }
		int rv=ftdi_read_data(&usb_context->ctx,buf2,2*size);
		if(verbose>3 /*&& rv!=0*/)
		{  fprintf(stderr,/*"ftdi_read_data(%u nibble-bytes)="*/" %d\n",
			/*2*size,*/rv);  }
		if(rv<0)
		{  Error("read",rv,1,0);  return(-1);  }
		
		int read_nibbles=0;
		if(usb_context->read_nibble_carry)
		{
			buf[0]=usb_context->read_nibble_carry & 0x0fU;
			usb_context->read_nibble_carry=0U;
			++read_nibbles;
		}
		
		int i;
		for(i=0; i<rv; i++)
		{
			unsigned char read_tmp=buf2[i];
			unsigned char *dest=&buf[read_nibbles/2];
			if((read_nibbles++&1))  *dest|=read_tmp & 0xf0U;
			else  *dest=read_tmp>>4;
		}
		
		if((read_nibbles&1))
		{
			// Save carry. 
			usb_context->read_nibble_carry=0x80U|buf[read_nibbles/2];
			--read_nibbles;
		}
		
		if(verbose>3 && read_nibbles/2>0)
		{
			int i=0;
			fprintf(stderr,"USB_DoReadData:");
			for(i=0; i<read_nibbles/2; i++)
			{  fprintf(stderr," %02x",(unsigned int)buf[i]);  }
			fprintf(stderr,"\n");
		}
		
		return(read_nibbles/2);
	}
	else
	{
		if(verbose>3)
		{  fprintf(stderr,"ftdi_read_data(%u bytes)= ",size);  }
		int rv=ftdi_read_data(&usb_context->ctx,buf,size);
		if(verbose>3 /*&& rv!=0*/)
		{  fprintf(stderr,/*"ftdi_read_data(%u bytes)="*/" %d\n",
			/*size,*/rv);  }
		if(rv<0)
		{  Error("read",rv,1,0);  return(-1);  }
		
		return(rv);
	}
}


// Returns -1 on error. 
static int USB_PollRead_Internal(void)
{
	if(usb_context->reading_hdr_bytes>=0)
	{
		// Try to read packet header. 
		int rv=USB_DoReadData((unsigned char*)usb_context->recv_pack,
				USBPPACKET_HEADER_SIZE-usb_context->reading_hdr_bytes);
		if(rv<0)  return(-1);
		if(rv>0)
		{
			usb_context->reading_hdr_bytes+=rv;
			if(usb_context->reading_hdr_bytes>=USBPPACKET_HEADER_SIZE)
			{
				assert(usb_context->reading_hdr_bytes==USBPPACKET_HEADER_SIZE);
				// Okay, header here. 
				if(verbose>1)
				{  fprintf(stderr,"RECV HDR: cmd=0x%x, len=%u\n",
					usb_context->recv_pack->command,
					usb_context->recv_pack->length);  }
				
				if(usb_context->recv_pack->length<USBPPACKET_HEADER_SIZE)
				{
					fprintf(stderr,"Received illegal packet size (%d).\n",
						(int)usb_context->recv_pack->length);
					// Stay in reading_hdr mode. 
					usb_context->reading_hdr_bytes=0;
				}
				else
				{
					usb_context->body_read_done=usb_context->reading_hdr_bytes;
					usb_context->reading_hdr_bytes=-1;
				}
			}
		}
	}
	if(usb_context->reading_hdr_bytes<0)  // no else!
	{
		unsigned char *base=((unsigned char*)usb_context->recv_pack);
		int rv=USB_DoReadData(base+usb_context->body_read_done,
				usb_context->recv_pack->length-usb_context->body_read_done);
		if(rv<0)  return(-1);
		usb_context->body_read_done+=rv;
		if(usb_context->body_read_done>=usb_context->recv_pack->length)
		{
			// Got packet completely. 
			usb_context->reading_hdr_bytes=0;
			usb_context->body_read_done=0;
			
			if(verbose>1)
			{  fprintf(stderr,"RECV: packet complete (cmd=0x%x, len=%u)\n",
				usb_context->recv_pack->command,usb_context->recv_pack->length);  }
			
			// We received something. Return status. 
			return(1);
		}
	}	
	
	return(0);
}


int USB_PollRead(void)
{
	int rv=USB_PollRead_Internal();
	return(rv>0 ? usb_context->recv_pack->length : rv);
}


// NOTE: expect_size without the 2 bytes for the header. 
// Return value: 
//  USB_OK -> OK
//  USB_Timeout -> timeout
//  USB_Interrupted -> SIGINT
//  USB_Unexpected -> wrong type or size
static USB_Status USB_WaitForReply(uint8 ptype,uint8 expect_size)
{
	expect_size+=USBPPACKET_HEADER_SIZE;
	
	struct timespec delay;
	delay.tv_sec=0;
	delay.tv_nsec=busy_poll_usec*1000;
	
	// timev = current time + timeout. 
	struct timeval timev,curr;
	gettimeofday(&timev,NULL);
	timev.tv_sec+= usb_resp_timeout/1000;
	timev.tv_usec+=usb_resp_timeout%1000;
	timev.tv_sec+=timev.tv_usec/1000000;
	timev.tv_usec%=1000000;
	// Poll USB: 
	for(;;)
	{
		int rv=USB_PollRead();
		if(rv<0)  exit(1);
		if(rv)  break;
		
		// We may be interrupted by a signal here but in this case, the 
		// polling interval will just be smaller this time. This does not 
		// matter much (as long as we are not bombarded with signals). 
		//usleep(busy_poll_usec);
		if(busy_poll_usec)  nanosleep(&delay,NULL);
		if(caught_sigint)  return(USB_Interrupted);
		
		// Check timeout. 
		gettimeofday(&curr,NULL);
		if(curr.tv_sec>timev.tv_sec || 
		   (curr.tv_sec==timev.tv_sec && curr.tv_usec>timev.tv_usec))
		{  return(USB_Timeout);  }
	}
	
	if(usb_context->recv_pack->command!=ptype || 
		usb_context->recv_pack->length!=expect_size)
	{
		if(usb_context->recv_pack->command==USB_CMD_ERROR)
		{  return(USB_ErrorReply);  }
		return(USB_Unexpected);
	}
	
	return(USB_OK);
}


// Send an identify request. Will not wait. 
static void USB_SendIdentifyRequest(void)
{
	// Send IDENTIFY request. 
	struct USBSendPacket p;
	p.command=USB_CMD_IDENTIFY;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_ident_req);
	
	if(verbose>2)
	{  fprintf(stderr,"sending identify request (%u bytes)\n",
		p.length);  }
	
	USB_DoWriteData((unsigned char*)&p,p.length);
}


USB_Status USB_GetBoardIdentification(USBIdentReply *ident)
{
	USB_SendIdentifyRequest();
	
	USB_Status rv=USB_WaitForReply(USB_CMD_IDENTIFY,
		sizeof(struct USBPacket_ident_resp));
	if(rv)  return(rv);
	
	memset(ident,0,sizeof(USBIdentReply));
	
	struct USBRecvPacket *p=usb_context->recv_pack;
	memcpy(ident->device_name,p->ident_resp.device_name,
		sizeof(p->ident_resp.device_name));
	memcpy(ident->device_vendor,p->ident_resp.device_vendor,
		sizeof(p->ident_resp.device_vendor));
	ident->firmware_version=p->ident_resp.firmware_version;
	ident->protocol_version=p->ident_resp.protocol_version;
	ident->atype=p->ident_resp.adapter_type;
	
	// First, check adapter type. 
	switch(ident->atype)
	{
		case 0:  usb_context->atype=AT_AtmelPrg;  break;
		case 1:  usb_context->atype=AT_UniProg;   break;
		default: usb_context->atype=AT_None;
	}
	if(usb_context->atype==AT_None)
	{  return(USB_UnknownBoard);  }
	
	// We currently accept firmware version 0x0202, only. 
	// We need protocol version >=3 for the IO speed command. 
	// We need protocol version >=4 for waveform command. 
	// Protocol 5 has different status request (to control the D6,7,8 pins). 
	int board_known = ident->firmware_version==0x0202 && 
	                  ident->protocol_version==5;
	// Minor changes to 0x0203 (waveform mode 1 allows delay). 
	board_known |= ident->firmware_version==0x0203 && 
	               ident->protocol_version==5;
	// Firmware version 2.1 is the same as 2.0, just some more inline 
	// assembly and more accurate timings. 
	// Firmware 2.2 has changed waveform command. 
	//board_known |= ident->firmware_version==0x0201 && 
	//               ident->protocol_version==4;
	if(!board_known)
	{  return(USB_UnknownBoard);  }
	
	return(USB_OK);
}


// set_led: 0 off, 1 on, or -1 for no change. 
static int _USB_SendStatusRequest(uint8 set_led,uint8 atmel_reset,
	uint8 spi_status,uint8 sd6,uint8 sd7,uint8 sd8)
{
	// Send STATUS request. 
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_STATUS;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_status_req);
	
	p.uap_status_req.set_led=set_led;
	p.uap_status_req.atmel_reset=atmel_reset;
	p.uap_status_req.spi_status=spi_status;
	p.uap_status_req.sd6=sd6;
	p.uap_status_req.sd7=sd7;
	p.uap_status_req.sd8=sd8;
	
	if(verbose>2)
	{  fprintf(stderr,"sending status request (%u bytes) (0x%02x,0x%02x)\n",
		p.length,p.uap_status_req.atmel_reset,spi_status);  }
	return(USB_DoWriteData((unsigned char*)&p,p.length));
}


static APinState _TranslatePinState(uint8 val,uint8 dir,uint8 in,
	int bit)
{
	val = (val>>bit) & 1U;
	dir = (dir>>bit) & 1U;
	in  = (in>>bit) & 1U;
	
	if(dir)
	{
		// Dir is set, i.e. it is output. 
		if(val)  return(APS_Out_1);
		else  return(APS_Out_0);
	}
	else if(val)
	{
		// Dir is not set, i.e. input; val is set, i.e. pull-ups. 
		if(in)  return(APS_PullUp_1);
		else  return(APS_PullUp_0);
	}
	else
	{
		if(in)  return(APS_HighZ_1);
		else  return(APS_HighZ_0);
	}
}


static void _USB_TranslateProgrammerStatus(USBProgrammerStatus *stat)
{
	struct USBRecvPacket *p=usb_context->recv_pack;
	
	struct USBPacket_uap_status_resp *pr=&p->uap_status_resp;
	stat->sck_state=  _TranslatePinState(pr->portc_val,pr->portc_dir,pr->portc_in,0);
	stat->mosi_state= _TranslatePinState(pr->portc_val,pr->portc_dir,pr->portc_in,1);
	stat->reset_state=_TranslatePinState(pr->portc_val,pr->portc_dir,pr->portc_in,2);
	stat->d6_state=   _TranslatePinState(pr->portc_val,pr->portc_dir,pr->portc_in,3);
	stat->d7_state=   _TranslatePinState(pr->portc_val,pr->portc_dir,pr->portc_in,4);
	stat->d8_state=   _TranslatePinState(pr->portc_val,pr->portc_dir,pr->portc_in,5);
	
	stat->led_is_on=
		(p->uap_status_resp.b_status & UAP_B_STATUS_LED_ON) ? 1 : 0;
	stat->miso_is_high=
		(p->uap_status_resp.b_status & UAP_B_STATUS_MISO_HIGH) ? 1 : 0;
	stat->led_is_busy_indicator=
		(p->uap_status_resp.b_status & UAP_B_STATUS_BUSY_LED_ENABLED) ? 1 : 0;
}


static inline uint8 _USB_ConvertSDPinValue(char v)
{
	switch(v)
	{
		// DDRC = 0x01, PORTC = 0x02. 
		case 'l':  return(0x80U | 0x01U);
		case 'h':  return(0x80U | 0x03U);
		case 'p':  return(0x80U | 0x02U);
		case 'z':  return(0x80U | 0x00U);
		// case 0 and default: fall off. 
	}
	return(0x00U);
}

// reset_val: 'l'ow, 'h'high, 'p'ullups, high-'z'
// spi_val: 'e'nable, 'p'ullups, high-'z'
// sd6,7,8: 'l'ow, 'h'high, 'p'ullups, high-'z'
USB_Status USB_SetPinState(char _spi_val,char _reset_val,
	char sd6_val,char sd7_val,char sd8_val,USBProgrammerStatus *stat)
{
	uint8 spi_val=0U;
	switch(_spi_val)
	{
		case 'z':  spi_val = 0x80U | 0x00U;  break;
		case 'e':  spi_val = 0x80U | 0x01U;  break;
		case 'p':  spi_val = 0x80U | 0x02U;  break;
		// default: keep 0U
	}
	
	uint8 res_val=0U;
	// 0x40: Change reset after SPI. 
	switch(_reset_val)
	{
		case 'l':  res_val = 0x40U | 0x00U;  break;
		case 'h':  res_val = 0x40U | 0x01U;  break;
		case 'z':  res_val = 0x40U | 0x02U;  break;
		case 'p':  res_val = 0x40U | 0x03U;  break;
		// default: keep 0U
	}
	
	uint8 sd6,sd7,sd8;
	sd6=_USB_ConvertSDPinValue(sd6_val);
	sd7=_USB_ConvertSDPinValue(sd7_val);
	sd8=_USB_ConvertSDPinValue(sd8_val);
	
	if(_USB_SendStatusRequest(0xff,res_val,spi_val,sd6,sd7,sd8))
	{  return(USB_WriteFailed);  }
	
	USB_Status rv=USB_WaitForReply(USB_CMD_UAP_STATUS,
		sizeof(struct USBPacket_uap_status_resp));
	if(rv)  return(rv);
	
	if(stat)
	{  _USB_TranslateProgrammerStatus(stat);  }
	
	return(USB_OK);
}


void USB_InitUniProgIOState(USBUniProgIOState *ios)
{
	ios->set_led=0xff;
	ios->atmel_reset=0x00;
	
	ios->ctrl_mask=0x00;
	ios->ddr_bits=0x00;
	ios->level_bits=0x00;
	
	ios->set_relais=0x00;
	
	ios->io_voltage=0xffffU;
	ios->sup_voltage=0xffffU;
	
	ios->bb_ddr=0;
	ios->bb_level=0;
	ios->bb_in=0;
	
	ios->relais_dio_closed=0;
	ios->relais_5v_closed=0;
	ios->relais_ldo_closed=0;
	ios->relais_gnd_closed=0;
	ios->ldo_vccio_enabled=0;
	ios->ldo_sup_enabled=0;
	ios->led_on=0;
	ios->busy_led_enabled=0;
	
	int i;
	for(i=0; i<4; i++)
	{  ios->adc_val[i]=0;  }
}


USB_Status USB_SetIOState(USBUniProgIOState *ios)
{
	// Send IOSTATUS request. 
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_IOSTATUS;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_iostatus_req);
	
	p.uap_iostatus_req.set_led=ios->set_led;
	
	uint8 tmp;
	switch(ios->atmel_reset)
	{
		case 'L':  tmp = 0 | (1U<<7);  break;
		case 'H':  tmp = 1 | (1U<<7);  break;
		case 'Z':  tmp = 2 | (1U<<7);  break;
		case 'l':  tmp = 0 | (1U<<6);  break;
		case 'h':  tmp = 1 | (1U<<6);  break;
		case 'z':  tmp = 2 | (1U<<6);  break;
		default: tmp=0;
	}
	p.uap_iostatus_req.atmel_reset=tmp;
	
	p.uap_iostatus_req.ctrl_mask=ios->ctrl_mask;
	p.uap_iostatus_req.ddr_bits=ios->ddr_bits;
	p.uap_iostatus_req.level_bits=ios->level_bits;
	p.uap_iostatus_req.set_relais=ios->set_relais;
	
	p.uap_iostatus_req.io_voltage=ios->io_voltage;
	p.uap_iostatus_req.sup_voltage=ios->sup_voltage;
	
	if(verbose>2)
	{  fprintf(stderr,"sending status request (%u bytes)\n",p.length);  }
	if(USB_DoWriteData((unsigned char*)&p,p.length))
	{  return(USB_WriteFailed);  }
	
	// ------
	
	USB_Status rv=USB_WaitForReply(USB_CMD_UAP_IOSTATUS,
		sizeof(struct USBPacket_uap_iostatus_resp));
	if(rv)  return(rv);
	
	// Fill in return values. 
	{
		struct USBRecvPacket *p=usb_context->recv_pack;
		struct USBPacket_uap_iostatus_resp *pr=&p->uap_iostatus_resp;
		
		ios->bb_ddr=pr->bb_ddr;
		ios->bb_level=pr->bb_level;
		ios->bb_in=pr->bb_in;
		
		ios->relais_dio_closed = (pr->status & UAP_IOSTAT_RELAIS_DIO_CLOSED) ? 1 : 0;
		ios->relais_5v_closed =  (pr->status & UAP_IOSTAT_RELAIS_5V_CLOSED) ? 1 : 0;
		ios->relais_ldo_closed = (pr->status & UAP_IOSTAT_RELAIS_LDO_CLOSED) ? 1 : 0;
		ios->relais_gnd_closed = (pr->status & UAP_IOSTAT_RELAIS_GND_CLOSED) ? 1 : 0;
		ios->ldo_vccio_enabled = (pr->status & UAP_IOSTAT_LDO_VCCIO_ENABLED) ? 1 : 0;
		ios->ldo_sup_enabled =   (pr->status & UAP_IOSTAT_LDO_SUP_ENABLED) ? 1 : 0;
		ios->led_on =            (pr->status & UAP_IOSTAT_LED_ON) ? 1 : 0;
		ios->busy_led_enabled =  (pr->status & UAP_IOSTAT_BUSY_LED_ENABLED) ? 1 : 0;
		
		int i;
		for(i=0; i<4; i++)
		{  ios->adc_val[i]=pr->adc_val[i];  }
	}
	
	return(USB_OK);
}


static USB_Status _USB_AtmelSimpleCmd(uint8 cmd,uint32 v32,uint8 v8)
{
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_ATCMD;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_atcmd_req);
	
	p.uap_atcmd_req.cmd=cmd;
	p.uap_atcmd_req.v8=v8;
	p.uap_atcmd_req.v32=v32;
	
	if(verbose>2)
	{  fprintf(stderr,"sending atcmd request (%u bytes)\n",
		p.length);  }
	if(USB_DoWriteData((unsigned char*)&p,p.length))
	{  return(USB_WriteFailed);  }
	
	USB_Status rv=USB_WaitForReply(USB_CMD_UAP_ATCMD,
		sizeof(struct USBPacket_uap_atcmd_resp));
	if(rv)  return(rv);
	
	return(USB_OK);
}

USB_Status USB_AtmelConnect(uint8 try_clock_shift)
{
	// We try 2 times. 
	int _i=2;
	for(; _i; _i--)
	{
		USB_Status rv=_USB_AtmelSimpleCmd(UAP_ATCMD_ENTER_PROG_MODE,
			/*v32=*/0,/*v8=*/try_clock_shift);
		if(rv)  return(rv);
		
		struct USBRecvPacket *p=usb_context->recv_pack;
		if(p->uap_atcmd_resp.cmd!=UAP_ATCMD_ENTER_PROG_MODE)
		{  return(USB_Unexpected);  }
		if(p->uap_atcmd_resp.status && _i<=1)
		{
			fprintf(stderr,"Failed to connect to Atmel. (readback=0x%x)\n",
				p->uap_atcmd_resp.v32);
			return(USB_ConnectFailed);
		}
		
		// Try to set RESET HIGH and LOW before retrying. 
		// This is needed e.g. for AT90S2313. One can also make it manually 
		// using the commands "rh" "rl" "connect". 
		// FIXME
#warning "FIXME: Second quirk to get the device responding during connect."
		/*{
			USB_SetSPIPins(0,'h',NULL);
			USB_SetSPIPins(0,'l',NULL);
		}*/
	}
	
	return(USB_OK);
}


USB_Status USB_AtmelSetIOSpeed(uint8 avr_io_speed)
{
	USB_Status rv=_USB_AtmelSimpleCmd(UAP_ATCMD_SET_IO_SPEED,0,avr_io_speed);
	if(rv)  return(rv);
	
	struct USBRecvPacket *p=usb_context->recv_pack;
	if(p->uap_atcmd_resp.cmd!=UAP_ATCMD_SET_IO_SPEED)
	{  return(USB_Unexpected);  }
	
	// Output: status: actually active IO speed. 
	// Device may not refuse to set our speed value. 
	if(avr_io_speed!=0xffU && p->uap_atcmd_resp.status!=avr_io_speed)
	{  return(USB_Unexpected);  }
	
	return(USB_OK);
}


USB_Status USB_AtmelReadSignature(USBAtmelSignature *sig)
{
	USB_Status rv=_USB_AtmelSimpleCmd(UAP_ATCMD_READ_SIGNATURE,0,0);
	if(rv)  return(rv);
	
	struct USBRecvPacket *p=usb_context->recv_pack;
	if(p->uap_atcmd_resp.cmd!=UAP_ATCMD_READ_SIGNATURE || 
		p->uap_atcmd_resp.status)
	{  return(USB_Unexpected);  }
	
	//  Output: v32: vendor<<16 | flash_size<<8 | device
	uint32 v32=p->uap_atcmd_resp.v32;
	sig->vendor=    (v32>>16) & 0xffU;
	sig->flash_size=(v32>> 8) & 0xffU;
	sig->device=    (v32    ) & 0xffU;
	
	return(USB_OK);
}


USB_Status USB_AtmelExecSingleCommand(uint32 atmel_cmd,
	uint8 *retval)
{
	USB_Status rv=_USB_AtmelSimpleCmd(UAP_ATCMD_EXEC_SINGLE_CMD,atmel_cmd,0);
	if(rv)  return(rv);
	
	struct USBRecvPacket *p=usb_context->recv_pack;
	if(p->uap_atcmd_resp.cmd!=UAP_ATCMD_EXEC_SINGLE_CMD)
	{  return(USB_Unexpected);  }
	
	//  Output: status: read back status from _ATSerialIO(). 
	if(retval) *retval=p->uap_atcmd_resp.status;
	
	return(USB_OK);
}


USB_Status USB_ReadFlash(uint16 adr0,uint16 size,uint16 *dest0)
{
	const uint32 AT_CMD_ReadFlashL = 0x20000000U;
	const uint32 AT_CMD_ReadFlashH = 0x28000000U;
	const int pipeline_size = USB_GetPipelineSize();
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_READ2;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_read2_req);
	p.uap_read2_req.cmd_L=AT_CMD_ReadFlashL;
	p.uap_read2_req.cmd_H=AT_CMD_ReadFlashH;
	
	int in_pipeline=0;
	uint16 adr=adr0,end=adr0+size;
	// Sigint is caught when waiting for response and will result in error 
	// return. 
	int dump_counter=0;
	uint16 *dest=dest0;
	while((adr<end || in_pipeline) /*&& !caught_sigint*/)
	{
		uint16 need=end-adr;
		if(need>32)  need=32;  // Max length 32 (words). 
		
		if(need)
		{
			p.uap_read2_req.adr=adr;
			p.uap_read2_req.len=need;

			adr+=need;

			if(verbose>2)
			{  fprintf(stderr,"sending read2 request (%u bytes) (0x%x,%d)\n",
				p.length,p.uap_read2_req.adr,p.uap_read2_req.len);  }
			if(USB_DoWriteData((unsigned char*)&p,p.length))
			{  return(USB_WriteFailed);  }
			++in_pipeline;
		}
		
		// For pipelining, one must set >=2 or >=3 here (">=1" for sequential). 
		// Since version 2, the programmer does support pipelinig. 
		// It has an internal pipeline of 3 (large) packets. 
		if(in_pipeline>=pipeline_size || adr>=end)
		{
			USB_Status rv=USB_WaitForReply(USB_CMD_UAP_READ2,
				sizeof(struct USBPacket_uap_data_resp));
			if(rv)  return(rv);
			
			int left=dest0+size-dest;
			if(left>32)  left=32;  // Max length 32 (words). 
			
			struct USBRecvPacket *p=usb_context->recv_pack;
			uint16 i=0;
			for(; i<left; i++)
			{  *dest++=p->uap_data_resp.data16[i];  }
			--in_pipeline;
			
			if((++dump_counter)==5)
			{
				printf("%3d%%\b\b\b\b",
					(100*(int)(adr-adr0)+size/2)/(int)(size));
				fflush(stdout);
				dump_counter=0;
			}
		}
	}
	
	assert(dest0+size==dest);
	
	return(USB_OK);
}


USB_Status USB_WriteFlash(uint16 adr0,uint16 size,const uint16 *src,
	uint16 page_words)
{
	const uint32 AT_CMD_ReadFlashL  = 0x20000000U;
	const uint32 AT_CMD_ReadFlashH  = 0x28000000U;
	const uint32 AT_CMD_WriteFlashL = 0x40000000U;
	const uint32 AT_CMD_WriteFlashH = 0x48000000U;
	const uint32 AT_CMD_WritePage   = 0x4C000000U;
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_WRITE2;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_write2_req);
	p.uap_write2_req.cmd_WL=AT_CMD_WriteFlashL;
	p.uap_write2_req.cmd_WH=AT_CMD_WriteFlashH;
	
	if(page_words)
	{
		// Okay, we need to program page-by-page. 
		// FIXME: The current algo only works when starting at a page 
		//        start. 
		assert(adr0%page_words==0);
		
		uint16 adr=adr0,end=adr0+size;
		
		int dump_counter=0;
		// Sigint is caught when waiting for response and will result in error 
		// return. 
		for(;/*!caught_sigint*/;)
		{
			uint16 need=end-adr;
			if(!need)  break;
			
			if(need>32)  need=32;  // Packet limited to 32 words. 
			// Can only program one page at a time. 
			if(need>page_words)  need=page_words;
			
			p.uap_write2_req.adr=adr;
			p.uap_write2_req.len=need;
			
			// See if we need to write the page or if it is not yet filled. 
			if((adr+need)%page_words==0 || adr+need>=end)
			{  p.uap_write2_req.paging=2;  }
			else
			{  p.uap_write2_req.paging=1;  }
			p.uap_write2_req.cmd_WP = AT_CMD_WritePage | ((uint32)(adr)<<8);
			
			// Fill in data and make data polling "calculation": 
			uint16 i;
			uint8 found_poll_byte=0;
			for(i=0; i<need; i++)
			{
				uint16 val=*src++;
				p.uap_write2_req.data16[i]=val;
				
				if(found_poll_byte)  continue;
				
				// Okay, data polling is nice, of course, but different devices 
				// require different data polling algorithms: 
				// ATmega161 returns 0xff until done. 
				// AT90S2313 returns 0x80 during erase and 0x7f during write. 
				// So, for now, I use delay timing for these 3 values. 
				uint8 low=val&0xffU,high=val>>8;
				if(low!=0xffU && low!=0x7fU && low!=0x80U)
				{
					p.uap_write2_req.cmd_P = AT_CMD_ReadFlashL | 
						(((uint32)(adr+i))<<8);
					p.uap_write2_req.poll_byte=low;
					found_poll_byte=1;  continue;
				}
				if(high!=0xffU && high!=0x7fU && high!=0x80U)
				{
					p.uap_write2_req.cmd_P = AT_CMD_ReadFlashH | 
						(((uint32)(adr+i))<<8);
					p.uap_write2_req.poll_byte=high;
					found_poll_byte=1;  continue;
				}
			}
			if(!found_poll_byte)
			{
				p.uap_write2_req.cmd_P=0x0U;
				p.uap_write2_req.poll_byte=0xffU;   // <-- no polling
			}
			adr+=need;
			
			// Otherwise, delay timing as fallback: 
			p.uap_write2_req.delay=15U;
			
			if(verbose>2)
			{  fprintf(stderr,"sending write2 request (%u bytes) (0x%x,%d)\n",
				p.length,p.uap_write2_req.adr,p.uap_write2_req.len);  }
			if(USB_DoWriteData((unsigned char*)&p,p.length))
			{  return(USB_WriteFailed);  }
			
			USB_Status rv=USB_WaitForReply(USB_CMD_UAP_WRITE2,
				sizeof(struct USBPacket_uap_write_resp));
			if(rv)  return(rv);
			
			struct USBRecvPacket *rp=usb_context->recv_pack;
			if(rp->uap_write_resp.status)
			{
				fprintf(stderr,"Write failure: status=0x%02x\n",
					rp->uap_write_resp.status);
				return(USB_WriteFailed);
			}
			
			if((++dump_counter)==5)
			{
				printf("%3d%%\b\b\b\b",(100*(int)(adr-adr0)+size/2)/(int)(size));
				fflush(stdout);
				dump_counter=0;
			}
		}
	}
	else
	{
		// Program the flash linearly, not page-by-page. 
		
		uint16 adr=adr0,end=adr0+size;
		
		p.uap_write2_req.paging=0;
		p.uap_write2_req.delay=6;
		p.uap_write2_req.cmd_RL=AT_CMD_ReadFlashL;
		p.uap_write2_req.cmd_RH=AT_CMD_ReadFlashH;
		
		int dump_counter=0;
		// Sigint is caught when waiting for response and will result in error 
		// return. 
		for(;/*!caught_sigint*/;)
		{
			uint16 need=end-adr;
			if(!need)  break;
			if(need>32)  need=32;  // Packet limited to 32 words. 
			
			p.uap_write2_req.adr=adr;
			p.uap_write2_req.len=need;
			
			uint16 i;
			for(i=0; i<need; i++)
			{  p.uap_write2_req.data16[i]=*src++;  }
			
			adr+=need;
			
			if(verbose>2)
			{  fprintf(stderr,"sending write2 request (%u bytes) (0x%x,%d)\n",
				p.length,p.uap_write2_req.adr,p.uap_write2_req.len);  }
			if(USB_DoWriteData((unsigned char*)&p,p.length))
			{  return(USB_WriteFailed);  }
			
			USB_Status rv=USB_WaitForReply(USB_CMD_UAP_WRITE2,
				sizeof(struct USBPacket_uap_write_resp));
			if(rv)  return(rv);
			
			struct USBRecvPacket *rp=usb_context->recv_pack;
			if(rp->uap_write_resp.status)
			{
				fprintf(stderr,"Write failure: status=0x%02x\n",
					rp->uap_write_resp.status);
				return(USB_WriteFailed);
			}
			
			if((++dump_counter)==5)
			{
				printf("%3d%%\b\b\b\b",(100*(int)(adr-adr0)+size/2)/(int)(size));
				fflush(stdout);
				dump_counter=0;
			}
		}
	}
	
	return(USB_OK);
}


USB_Status USB_ReadEEPROM(uint16 adr0,uint16 size,uint8 *dest0)
{
	const uint32 AT_CMD_ReadEEPROM  = 0xA0000000U;
	const int pipeline_size = USB_GetPipelineSize();
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_READ;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_read_req);
	p.uap_read_req.cmd=AT_CMD_ReadEEPROM;
	
	int in_pipeline=0;
	uint16 adr=adr0,end=adr0+size;
	// Sigint is caught when waiting for response and will result in error 
	// return. 
	uint8 *dest=dest0;
	while((adr<end || in_pipeline) /*&& !caught_sigint*/)
	{
		uint16 need=end-adr;
		if(need>64)  need=64;  // Max length 64 (bytes). 
		
		if(need)
		{
			p.uap_read_req.adr=adr;
			p.uap_read_req.len=need;
			
			adr+=need;
			
			if(verbose>2)
			{  fprintf(stderr,"sending read request (%u bytes) (0x%x,%d)\n",
				p.length,p.uap_read_req.adr,p.uap_read_req.len);  }
			if(USB_DoWriteData((unsigned char*)&p,p.length))
			{  return(USB_WriteFailed);  }
			++in_pipeline;
		}
		
		// For pipelining, one must set >=2 or >=3 here (">=1" for sequential). 
		// Since version 2, the programmer does support pipelinig. 
		// It has an internal pipeline of 3 (large) packets. 
		if(in_pipeline>=pipeline_size || adr>=end)
		{
			USB_Status rv=USB_WaitForReply(USB_CMD_UAP_READ,
				sizeof(struct USBPacket_uap_data_resp));
			if(rv)  return(rv);
			
			int left=dest0+size-dest;
			if(left>64)  left=64;  // Max length 64 (bytes). 
			
			struct USBRecvPacket *p=usb_context->recv_pack;
			uint16 i=0;
			for(; i<left; i++)
			{  *dest++=p->uap_data_resp.data8[i];  }
			--in_pipeline;
			
			printf("%3d%%\b\b\b\b",(100*(int)(adr-adr0)+size/2)/(int)(size));
			fflush(stdout);
		}
	}
	
	assert(dest0+size==dest);
	
	return(USB_OK);
}


USB_Status USB_WriteEEPROM(uint16 adr0,uint16 size,const uint8 *src)
{
	const uint32 AT_CMD_ReadEEPROM  = 0xA0000000U;
	const uint32 AT_CMD_WriteEEPROM = 0xC0000000U;
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_WRITE;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_write_req);
	p.uap_write_req.cmd_W=AT_CMD_WriteEEPROM;
	p.uap_write_req.cmd_R=AT_CMD_ReadEEPROM;
	
	uint16 adr=adr0,end=adr0+size;
	
	// Sigint is caught when waiting for response and will result in error 
	// return. 
	for(;/*!caught_sigint*/;)
	{
		uint16 need=end-adr;
		if(!need)  break;
		if(need>64)  need=64;  // Packet limited to 64 bytes. 
		
		p.uap_write_req.adr=adr;
		p.uap_write_req.len=need;
		p.uap_write_req.delay=6;
		
		uint16 i;
		for(i=0; i<need; i++)
		{  p.uap_write_req.data8[i]=*src++;  }
		
		adr+=need;
		
		if(verbose>2)
		{  fprintf(stderr,"sending write request (%u bytes) (0x%x,%d)\n",
			p.length,p.uap_write_req.adr,p.uap_write_req.len);  }
		if(USB_DoWriteData((unsigned char*)&p,p.length))
		{  return(USB_WriteFailed);  }
		
		USB_Status rv=USB_WaitForReply(USB_CMD_UAP_WRITE,
			sizeof(struct USBPacket_uap_write_resp));
		if(rv)  return(rv);
		
		struct USBRecvPacket *rp=usb_context->recv_pack;
		if(rp->uap_write_resp.status)
		{
			fprintf(stderr,"Write failure: status=0x%02x\n",
				rp->uap_write_resp.status);
			return(USB_WriteFailed);
		}
		
		printf("%3d%%\b\b\b\b",(100*(int)(adr-adr0)+size/2)/(int)(size));
		fflush(stdout);
	}
	
	return(USB_OK);
}


USB_Status USB_WaveformIO(uint8 wave_mode,int wave_size,uint8 *wave_data,
	uint8 wave_delay)
{
	switch(wave_mode)
	{
		case 0:  assert(wave_size<=128);  break;
		case 1:  assert(wave_size<=256);  break;
		default:  assert(0);
	}
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_WAVEFORM;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_waveform_req);
	
	memset(&p.uap_waveform_req,0,sizeof(struct USBPacket_uap_waveform_req));
	p.uap_waveform_req.wave_mode=wave_mode;
	p.uap_waveform_req.delay=wave_delay;
	p.uap_waveform_req.wave_size=wave_size;
	int i;
	switch(wave_mode)
	{
		case 0:
			// NOTE: As input, we expect wave_data to hold wave_size bytes but 
			//       we "compress" this keeping only the lowest 4 bits of the 
			//       wave. 
			for(i=0; i<wave_size; i++)
			{
				if(i&0x1)  // i is odd
				{  p.uap_waveform_req.wave_data[i/2]|=wave_data[i]<<4;  }
				else  // i is even
				{  p.uap_waveform_req.wave_data[i/2]=wave_data[i]&0x0fU;  }
			}
			break;
		case 1:
			// NOTE: As input, we expect wave_data to hold wave_size bytes but 
			//       we "compress" this keeping only the lowest 2 bits of the 
			//       wave. 
			for(i=0;;)
			{
				uint8 *d=&(p.uap_waveform_req.wave_data[i/4]);
				if(i>=wave_size) break;  *d=  wave_data[i++]&0x03U;
				if(i>=wave_size) break;  *d|=(wave_data[i++]&0x03U)<<2;
				if(i>=wave_size) break;  *d|=(wave_data[i++]&0x03U)<<4;
				if(i>=wave_size) break;  *d|=(wave_data[i++]&0x03U)<<6;
			}
			break;
		default:  assert(0);
	}
	
	if(verbose>2)
	{  fprintf(stderr,"sending waveform request (%u bytes, wave_size=%u)\n",
		p.length,(unsigned int)p.uap_waveform_req.wave_size);  }
	if(USB_DoWriteData((unsigned char*)&p,p.length))
	{  return(USB_WriteFailed);  }
	
	USB_Status rv=USB_WaitForReply(USB_CMD_UAP_WAVEFORM,
		sizeof(struct USBPacket_uap_waveform_resp));
	if(rv)  return(rv);
	
	struct USBPacket_uap_waveform_resp *resp=&usb_context->recv_pack->uap_waveform_resp;
	if(verbose>2 || (int)resp->resp_bits!=wave_size)
	{
		const uint8 *data=(uint8*)usb_context->recv_pack;
		fprintf(stderr,"got waveform response (A) "
			"(resp_bits=%u=0x%x; wave_size=%u=0x%x  %02x %02x %02x %02x %02x %02x)\n",
			(unsigned int)resp->resp_bits,(unsigned int)resp->resp_bits,
			(unsigned int)wave_size,(unsigned int)wave_size,
			data[0],data[1],data[2],data[3],data[4],data[5]);
	}
	assert(resp->resp_bits==wave_size);
	
	// Store the returned state in wave_data. 
	for(i=0; i<wave_size; i++)
	{
		// The bit we read in last is the LSB in miso_data. 
		wave_data[i] = (resp->miso_data[i/8]>>(7-(i&0x7))) & 0x01U;
		//wave_data[i] = resp->miso_data[i/8];
	}
	
	return(USB_OK);
}


static USB_Status _USB_RawWriteWaveformReq_AtmelPrg(uint8 wave_mode,
	int wave_size,const uint8 *wave_data_packed,uint8 wave_delay)
{
	switch(wave_mode)
	{
		case 0:  assert(wave_size<=128);  break;
		case 1:  assert(wave_size<=256);  break;
		default:  assert(0);
	}
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_WAVEFORM;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_waveform_req);
	
	memset(&p.uap_waveform_req,0,sizeof(struct USBPacket_uap_waveform_req));
	p.uap_waveform_req.wave_mode=wave_mode;
	p.uap_waveform_req.delay=wave_delay;
	p.uap_waveform_req.wave_size=wave_size;
	switch(wave_mode)
	{
		case 0:
			memcpy(p.uap_waveform_req.wave_data,wave_data_packed,
				(wave_size+1)/2);
			break;
		case 1:
			memcpy(p.uap_waveform_req.wave_data,wave_data_packed,
				(wave_size+3)/4);
			break;
		default:  assert(0);
	}
	
	if(verbose>2)
	{  fprintf(stderr,"sending waveform request (%u bytes, wave_size=%u)\n",
		p.length,(unsigned int)p.uap_waveform_req.wave_size);  }
	if(USB_DoWriteData((unsigned char*)&p,p.length))
	{  return(USB_WriteFailed);  }
	
	return(USB_OK);
}

static USB_Status _USB_RawWriteWaveformReq_UniProg(uint8 wave_mode,
	int wave_size,const uint8 *wave_data_packed,uint8 wave_delay)
{
	// mode=0 currently not supported by UniProg. UniProg will send error reply.
	switch(wave_mode)
	{
		case 0:  assert(wave_size<=128);  break;
		case 1:  assert(wave_size<=256);  break;
		default:  assert(0);
	}
	
	struct USBSendPacket p;
	p.command=USB_CMD_UAP_WAVEFORM;
	p.length=USBPPACKET_HEADER_SIZE+
		sizeof(struct USBPacket_uap_waveform_req);
	
	memset(&p.uap_waveform_req,0,sizeof(struct USBPacket_uap_waveform_req));
	p.uap_waveform_req.wave_mode=wave_mode;
	p.uap_waveform_req.delay=wave_delay;
	p.uap_waveform_req.wave_size=wave_size;
	switch(wave_mode)
	{
		case 0:
			memcpy(p.uap_waveform_req.wave_data,wave_data_packed,
				(wave_size+1)/2);
			break;
		case 1:
		{
			// Need to re-order the bits. 
			// In wave_data_packed bits 0,1 are MOSI,RESET and shifted out 
			// first. 
			// Translate this to 2 byte streams, even bytes MOSI, odd bytes 
			// RESET and bit 7 shifted out first. 
			//              [0] 7             0  [1] 7             0
			// Hence, 16 bits   H G F E D C B A      P O N M L K J I 
			// translated into  A C E G I K M O      B D F H J L N P
#if 0  /* Original code; working. */
			int i; 
			for(i=0; i<wave_size; i++)
			{
				uint8 msi = (wave_data_packed[i/4]>>(2*(i%4)  )) & 1U;
				uint8 rst = (wave_data_packed[i/4]>>(2*(i%4)+1)) & 1U;
				p.uap_waveform_req.wave_data[2*(i/8)  ] |= msi<<(7-(i%8));
				p.uap_waveform_req.wave_data[2*(i/8)+1] |= rst<<(7-(i%8));
			}
#else
			// FIXME: Optimize this code!
			int i; 
			for(i=0; i<wave_size; i++)
			{
				uint8 rst_msi = (wave_data_packed[i/4]>>(2*(i%4))) & 0x3U;
				int shft = 7-(i%8);
				int idx = 2*(i/8);
				p.uap_waveform_req.wave_data[idx  ] |= (rst_msi&1U)<<shft;
				p.uap_waveform_req.wave_data[idx+1] |= (rst_msi>>1)<<shft;
			}
#endif
		}	break;
		default:  assert(0);
	}
	
	if(verbose>2)
	{  fprintf(stderr,"sending waveform request (%u bytes, wave_size=%u)\n",
		p.length,(unsigned int)p.uap_waveform_req.wave_size);  }
	if(USB_DoWriteData((unsigned char*)&p,p.length))
	{  return(USB_WriteFailed);  }
	
	return(USB_OK);
}

USB_Status USB_RawWriteWaveformReq(uint8 wave_mode,int wave_size,
	const uint8 *wave_data_packed,uint8 wave_delay)
{
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
			return(_USB_RawWriteWaveformReq_AtmelPrg(wave_mode,wave_size,
				wave_data_packed,wave_delay));
		case AT_UniProg:
			return(_USB_RawWriteWaveformReq_UniProg(wave_mode,wave_size,
				wave_data_packed,wave_delay));
		default:  assert(0);
	}
	
	// This should never be reached. 
	return(USB_UnknownBoard);
}


USB_Status USB_RawReadWaveformResp(int wave_size,uint8 *wave_resp_packed)
{
	USB_Status rv=USB_WaitForReply(USB_CMD_UAP_WAVEFORM,
		sizeof(struct USBPacket_uap_waveform_resp));
	if(rv)  return(rv);
	
	struct USBPacket_uap_waveform_resp *resp=&usb_context->recv_pack->uap_waveform_resp;
	if(verbose>2 || (int)resp->resp_bits!=wave_size)
	{
		const uint8 *data=(uint8*)usb_context->recv_pack;
		fprintf(stderr,"got waveform response (B) "
			"(resp_bits=%u=0x%x; wave_size=%u=0x%x  %02x %02x %02x %02x %02x %02x)\n",
			(unsigned int)resp->resp_bits,(unsigned int)resp->resp_bits,
			(unsigned int)wave_size,(unsigned int)wave_size,
			data[0],data[1],data[2],data[3],data[4],data[5]);
	}
	assert(resp->resp_bits==wave_size);
	
	// Copy result: 
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
			memcpy(wave_resp_packed,resp->miso_data,(wave_size+7)/8);
			break;
		case AT_UniProg:
		{
			// Need to "unpack" the response: Swap bits. FIXME
			memcpy(wave_resp_packed,resp->miso_data,(wave_size+7)/8);
		}	break;
		default:  assert(0);
	}
	
	return(USB_OK);
}


USB_Status USB_RawWaveformIO(uint8 wave_mode,int wave_size,
	const uint8 *wave_data_packed,uint8 *wave_resp_packed,uint8 wave_delay)
{
	USB_Status rv=USB_RawWriteWaveformReq(wave_mode,wave_size,
		wave_data_packed,wave_delay);
	if(rv)  return(rv);
	
	return(USB_RawReadWaveformResp(wave_size,wave_resp_packed));
}


void USB_Reset(void)
{
	int rv=ftdi_usb_purge_buffers(&usb_context->ctx);
	Error("purge buffers",rv,1,1);
	
	rv=ftdi_usb_reset(&usb_context->ctx);
	Error("reset",rv,1,1);
	
	rv=ftdi_usb_purge_buffers(&usb_context->ctx);
	Error("purge buffers",rv,1,1);
	
	usb_context->reading_hdr_bytes=0;  // -1 -> body
	usb_context->body_read_done=0;
	usb_context->read_nibble_carry=0;
	
	// Write 4 dummy NUL bytes. 
	char buf[4]={0,0,0,0};
	USB_DoWriteData((unsigned char*)buf,4);
	
	usleep(10);
}


int USB_SetLatency(int msec)
{
	if(msec>255) msec=255;
	else if(msec<1) msec=1;
	int rv=ftdi_set_latency_timer(&usb_context->ctx,msec);
	Error("latency",rv,1,1);
	return(rv);
}

int USB_GetLatency(void)
{
	unsigned char latency;
	int rv=ftdi_get_latency_timer(&usb_context->ctx,&latency);
	Error("latency(get)",rv,1,1);
	return((int)latency);
}


const char *USB_GetIdfString(void)
{
	return(usb_context->usb_idf_str);
}

AdapterType USB_GetAdapterType(void)
{
	return(usb_context->atype);
}

int USB_GetPipelineSize(void)
{
	switch(usb_context->atype)
	{
		case AT_AtmelPrg:  return(USB_LARGE_PIPELINE_SIZE_ATMELPRG);
		case AT_UniProg:   return(USB_LARGE_PIPELINE_SIZE_UNIPROG);
		default:  assert(0);
	}
	return(-1);
}

struct ftdi_eeprom *USB_ReadFTDIEEPROM()
{
	unsigned char buf[128];
	int rv=ftdi_read_eeprom(&usb_context->ctx,buf);
	Error("read eeprom",rv,1,1);
	struct ftdi_eeprom *eep=(struct ftdi_eeprom *)malloc(
		sizeof(struct ftdi_eeprom));
	ftdi_eeprom_unbuild(buf,eep);
	return(eep);
}

void USB_FreeFTDIEEPROM(struct ftdi_eeprom *eep)
{
	if(eep)
	{
		ftdi_eeprom_free(eep);
		free(eep);
	}
}


// Note: Pass eep=NULL to just erase the EEPROM. 
int USB_WriteFTDIEEPROM(struct ftdi_eeprom *eep)
{
	unsigned char buf[128];
	if(eep)
	{
		int rv=ftdi_eeprom_build(eep,buf);
		if(rv<0)
		{
			fprintf(stderr,"Failed to compose EEPROM image; probably some "
				"strings too long.\n");
			return(1);
		}
	}
	
	int rv=ftdi_erase_eeprom(&usb_context->ctx);
	Error("erase eeprom",rv,1,1);
	if(eep)
	{
		rv=ftdi_write_eeprom(&usb_context->ctx,buf);
		Error("write eeprom",rv,1,1);
	}
	
	return(0);
}


void USB_OpenConection(int vendor,int device)
{
	assert(!usb_context);
	usb_context = (USBContext*)malloc(sizeof(USBContext));
	if(!usb_context)
	{  fprintf(stderr,"Allocation failure\n");  abort();  }
	
	usb_context->ctx_valid=0;
	usb_context->atype=AT_None;
	usb_context->enable_nibble_io=2;
	usb_context->read_nibble_carry=0;
	usb_context->reading_hdr_bytes=0;
	usb_context->body_read_done=0;
	usb_context->recv_pack=(struct USBRecvPacket*)&usb_context->usb_recv_buf;
	
	int rv=ftdi_init(&usb_context->ctx);
	Error("init",rv,0,1);
	usb_context->ctx_valid=1;
	
	if(verbose)
	{  fprintf(stderr,"Using vendor=0x%04x, device=0x%04x...\n",
		vendor,device);  }
	rv=ftdi_usb_open(&usb_context->ctx,vendor,device);
	if(rv)
	{  fprintf(stderr,
		"Failed to open USB connection (vendor=0x%04x, device=0x%04x):\n",
		vendor,device);  }
	Error("open",rv,0,1);
	usb_context->ctx_valid=2;
	
	rv=ftdi_usb_reset(&usb_context->ctx);
	Error("reset",rv,1,1);
	
	rv=ftdi_usb_purge_buffers(&usb_context->ctx);
	Error("purge buffers",rv,1,1);
	
	// Be sure with bitbang mode: 
	rv=ftdi_disable_bitbang(&usb_context->ctx);
	Error("bitbang(off)",rv,1,1);
	
	unsigned char latency;
	rv=ftdi_get_latency_timer(&usb_context->ctx,&latency);
	Error("latency(get)",rv,1,1);
	//fprintf(stderr,"latency=%u  (0..255 ms)\n",latency);
	
	// Since the ATMega8 needs about 1.27msec to transfer a 64byte buffer 
	// into the FTDI SIO (8MHz version), a latency of 2msec should be perfect. 
	//rv=ftdi_set_latency_timer(&usb_context->ctx,/*msec=*/2);
	// It turns out experimentally (with pipelining) that a latency of 3 is 
	// equally fast for normal speed but faster for "fast" and "extreme" speed. 
	rv=ftdi_set_latency_timer(&usb_context->ctx,/*msec=*/3);
	Error("latency",rv,1,1);
	
	unsigned int read_data_chunksize;
	rv=ftdi_read_data_get_chunksize(&usb_context->ctx,&read_data_chunksize);
	Error("read_data_chunksize(get)",rv,1,1);
	//fprintf(stderr,"read_data_chunksize=%u\n",read_data_chunksize);
	
	// The FIFO has a 384 byte fifo buffer; using more has no point here. 
	rv=ftdi_read_data_set_chunksize(&usb_context->ctx,384);
	Error("read_data_chunksize(set)",rv,1,1);
	
	rv=ftdi_read_data_get_chunksize(&usb_context->ctx,&read_data_chunksize);
	Error("read_data_chunksize(get)",rv,1,1);
	//fprintf(stderr,"read_data_chunksize=%u\n",read_data_chunksize);
	
	// Read info from the EEPROM attached to the FTDI SIO: 
	struct ftdi_eeprom *eep=USB_ReadFTDIEEPROM();
	// We need to decide NOW whether we need nibble IO or not. 
	if(eep->product)
	{
		if(strcasestr(eep->product,"uniprog"))
		{  usb_context->enable_nibble_io=0;  }
		else if(strcasestr(eep->product,"atmelprg"))
		{  usb_context->enable_nibble_io=1;  }
		if(strcasestr(eep->product,"uni"))
		{  usb_context->enable_nibble_io=0;  }
		else if(strcasestr(eep->product,"atmel"))
		{  usb_context->enable_nibble_io=1;  }
	}
	if(usb_context->enable_nibble_io>1)
	{
		fprintf(stderr,"Device name not recognized; assuming AtmelPrg "
			"with nibble IO.\n");
		usb_context->enable_nibble_io=2;
	}
	snprintf(usb_context->usb_idf_str,256,
		"%04x:%04x (%s, %s)"
		,eep->vendor_id,eep->product_id,
		eep->product ? eep->product : "[no product name]",
		eep->serial ? eep->serial : "[no serial]");
	// Also tell the user: 
	printf(
		"Opened USB connection to %04x:%04x:\n"
		"  Hardware: %s, %s (%s)\n"
		,eep->vendor_id,eep->product_id,
		eep->product ? eep->product : "[no product name]",
		eep->serial ? eep->serial : "[no serial]",
		usb_context->enable_nibble_io ? "nibble IO" : "byte IO");
	USB_FreeFTDIEEPROM(eep);
	
	// Write some dummy NUL bytes and read everything that comes back. 
	// This helps us synchronize in the beginning. 
	{
		// Buffer length is 128 but only the first 4 bytes used for writing, 
		// the others for reading. 
		char buf[128]={0,0,0,0};
		USB_DoWriteData((unsigned char*)buf,4);
		usleep(10);
		
		USB_DoReadData(buf,128);
	}
}


void USB_CloseConnection(void)
{
	if(usb_context)
	{
		if(usb_context->ctx_valid>=2)
		{
			int rv=ftdi_usb_close(&usb_context->ctx);
			usb_context->ctx_valid=1;
			Error("close",rv,0,1);
		}
		if(usb_context->ctx_valid>=1)
		{
			ftdi_deinit(&usb_context->ctx);
			usb_context->ctx_valid=0;
		}
		
		free(usb_context);
		usb_context=NULL;
	}
}
