/*
 * usbio.h -- USB Atmel programmer USB access routines. 
 * 
 * Copyright (c) 2005--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef _USB_ATMEL_PRG_USBIO_H_
#define _USB_ATMEL_PRG_USBIO_H_ 1

#include <ftdi.h>
#include "prototypes.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum _USB_Status
{
	USB_OK=0,
	USB_Timeout,
	USB_Interrupted,
	USB_Unexpected,
	USB_ErrorReply,
	USB_UnknownBoard,
	USB_ConnectFailed,
	USB_WriteFailed,
}USB_Status;

// Support 2 different cable adapters. 
typedef enum _AdapterType
{
	AT_None=0,
	AT_AtmelPrg,
	AT_UniProg,
}AdapterType;

// Convert status to string. 
extern const char *USB_Status_String(USB_Status status);


struct _USBContext;  // Opaque. 
typedef struct _USBContext USBContext;

// Global pointer to USB context or NULL. 
extern USBContext *usb_context;

typedef struct _USBIdentReply
{
	char device_name[64];
	char device_vendor[64];
	uint16 firmware_version;
	uint8 protocol_version;
	AdapterType atype;
} USBIdentReply;


// Bit 0 is the pin value
typedef enum _APinState
{
	APS_HighZ_0=0x00,   // High-Z, reading 0
	APS_HighZ_1=0x01,   // High-Z, reading 1
	APS_PullUp_0=0x02,  // Pull-Up, reading 0
	APS_PullUp_1=0x03,  // Pull-Up, reading 0
	APS_Out_0=0x04,     // Output, 0
	APS_Out_1=0x05,     // Output, 1
} APinState;

typedef struct _USBProgrammerStatus
{
	APinState sck_state : 4;
	APinState mosi_state : 4;
	APinState reset_state : 4;
	APinState d6_state : 4;  // diagnose on port 6 of Sub-D
	APinState d7_state : 4;  // diagnose on port 7 of Sub-D
	APinState d8_state : 4;  // diagnose on port 8 of Sub-D
	
	int led_is_on : 1;
	int miso_is_high : 1;
	int led_is_busy_indicator : 1;
} USBProgrammerStatus;


typedef struct _USBAtmelSignature
{
	uint8 vendor;
	uint8 flash_size;
	uint8 device;
} USBAtmelSignature;


// NOTE: Same as in bitbang.h in firmware. 
#define UNIPROG_BIT_DDR_TCK_SCK   (1U<<0)
#define UNIPROG_BIT_DDR_TDO_MISO  (1U<<1)
#define UNIPROG_BIT_DDR_TDI_MOSI  (1U<<2)
#define UNIPROG_BIT_DDR_TMS_RESET (1U<<3)
#define UNIPROG_BIT_DDR_D6        (1U<<4)
#define UNIPROG_BIT_DDR_D7        (1U<<5)
#define UNIPROG_BIT_DDR_D8RES     (1U<<6)

// NOTE: Same as in usb_types.h in firmware. 
#define UAP_RELAIS_DIO_CLOSE  (1U<<0)
#define UAP_RELAIS_DIO_OPEN   (1U<<1)
#define UAP_RELAIS_5V_CLOSE   (1U<<2)
#define UAP_RELAIS_5V_OPEN    (1U<<3)
#define UAP_RELAIS_LDO_CLOSE  (1U<<4)
#define UAP_RELAIS_LDO_OPEN   (1U<<5)
#define UAP_RELAIS_GND_CLOSE  (1U<<6)
#define UAP_RELAIS_GND_OPEN   (1U<<7)

typedef struct _USBUniProgIOState
{
	// -------------INPUT PARAMETERS--------------
	// See USBPacket_uap_iostatus_req. 
	uint8 set_led;
	// Change reset line: 'l'ow, 'h'igh, high-'z'. 
	// Small letters: after IO change, capital letters before IO change. 
	char atmel_reset;
	// Control the IO pins. See USBPacket_uap_iostatus_req. 
	uint8 ctrl_mask;
	uint8 ddr_bits;
	uint8 level_bits;
	// Set relais. See USBPacket_uap_iostatus_req and UAP_RELAIS_*
	uint8 set_relais;
	// Choose IO and supply voltages. 150..500 in 0.01V. 
	// Use 0000 to shut down the regulator and 0xffff for "no change". 
	uint16 io_voltage;
	uint16 sup_voltage;
	
	// -------------OUTPUT PARAMETERS-------------
	// Bitbang port state. Note that when the JTAG system in the CPLD is 
	// active, some output ports can be overridden. 
	uint8 bb_ddr,bb_level,bb_in;
	// Relais status + LDO shutdown status (6 bits) + LED status (2 bits). 
	int relais_dio_closed : 1;
	int relais_5v_closed : 1;
	int relais_ldo_closed : 1;
	int relais_gnd_closed : 1;
	int ldo_vccio_enabled : 1;
	int ldo_sup_enabled : 1;
	int led_on : 1;
	int busy_led_enabled : 1;
	// Measured ADC voltages in mV. 
	uint16 adc_val[4];
} USBUniProgIOState;

// Report an USB error. Close if doclose is set; exit if doexit is set. 
void Error(const char *where,int rv,int doclose,int doexit);

// Open connection to the USB Atmel programmer device. 
// On failure, exits with error message written to stderr. 
// Global usb_context is non-NULL afterwards on success. 
// Currently quits on failure. 
void USB_OpenConection(int vendor,int device);

// Close a connection to the USB Atmel programmer device. 
// Do not use the global usb_context after that. 
void USB_CloseConnection(void);

// Call frequently to poll for incoming data. 
// Returns number of read bytes. Data discarted. 
// Returns -1 on error. 
int USB_PollRead(void);

// Reset communication. 
void USB_Reset(void);

// Get/Set latency value in msec (1..255). 
int USB_SetLatency(int msec);
int USB_GetLatency(void);

// Get USB identification string (what we're connected to). 
const char *USB_GetIdfString(void);

// Get adapter type. This just returns the value from the USB context. 
AdapterType USB_GetAdapterType(void);

// Get size of pipeline (number of large packets). 
int USB_GetPipelineSize(void);

// Get board ident data. 
// Returns USB_OK if okay, USB_UnknownBoard if unknown board; rest is errors. 
USB_Status USB_GetBoardIdentification(USBIdentReply *ident);

// Change reset pin to atmel to be programmed: 
//   reset_val, sd6,7,8: 'l'ow, 'h'high, high-'z', 'p'ull-ups. 
// Set pin states: spi_val: 'e'nable, 'p'ullups, high-'z'
// Can use spi_val=0, reset_val,sd6,7,8 = 0 to just query status. 
// May use stat=NULL. 
// ONLY FOR USB-AtmelPrg. Use USB_SetIOState() for USB-UniProg. 
USB_Status USB_SetPinState(char spi_val,char reset_val,
	char sd6_val,char sd7_val,char sd8_val,USBProgrammerStatus *stat);

// Set up the USBUniProgIOState to do nothing. 
void USB_InitUniProgIOState(USBUniProgIOState *ios);

// ONLY FOR USB-UniProg. 
USB_Status USB_SetIOState(USBUniProgIOState *ios);

// Connect programmer to atmel. 
USB_Status USB_AtmelConnect(uint8 try_clock_shift);

// Set speed for communication with device to be programmed. 
// Use 0 for extreme (0.5625usec), 
// Use 1 for fast, 2 for standard (2.125usec) and >=3 for delay timings. 
// 0xff for "no change" and 0 reserved for max speed. 
USB_Status USB_AtmelSetIOSpeed(uint8 avr_io_speed);

// Read signature. Only do after successful USB_AtmelConnect(). 
USB_Status USB_AtmelReadSignature(USBAtmelSignature *sig);

// Read flash content. Size in 2byte words. 
USB_Status USB_ReadFlash(uint16 adr,uint16 size,uint16 *dest);

// Write flash content. Size in 2byte words. 
USB_Status USB_WriteFlash(uint16 adr,uint16 size,
	const uint16 *src,uint16 page_words);

// Read EEPROM content. Size in bytes. 
USB_Status USB_ReadEEPROM(uint16 adr,uint16 size,uint8 *dest);

// Write EEPROM content. Size in bytes. 
USB_Status USB_WriteEEPROM(uint16 adr,uint16 size,
	const uint8 *src);

// Execute single atmel command. 
// The status readback is returned in *retval. 
USB_Status USB_AtmelExecSingleCommand(uint32 atmel_cmd,
	uint8 *retval);

// Send a single waveform request: Input data is passed in wave_data where 
// one byte corresponds to one state change and hence the wave_data[] array 
// has size wave_size. The function will "compress" the wave (2 or 4 bit 
// encoding depending on wave_mode), transmit it, wait for the result and 
// then store the result in wave_data again. Result wave_data will be of 
// size wave_size again; each read back bit is expanded into one byte of 
// wave_data being either 0 or 1. 
USB_Status USB_WaveformIO(uint8 wave_mode,int wave_size,
	uint8 *wave_data,uint8 wave_delay);

// "Raw" version of the waveform IO: Expects packed ("compressed") input 
// in wave_data_packed and will return read back output in wave_resp_packed. 
USB_Status USB_RawWaveformIO(uint8 wave_mode,int wave_size,
	const uint8 *wave_data_packed,uint8 *wave_resp_packed,uint8 wave_delay);

// These two functions are the send and receive part of the function 
// above separated to allow implementation of pipelining. 
USB_Status USB_RawWriteWaveformReq(uint8 wave_mode,int wave_size,
	const uint8 *wave_data_packed,uint8 wave_delay);
USB_Status USB_RawReadWaveformResp(int wave_size,uint8 *wave_resp_packed);

// Read FTDI config EEPROM content; returns allocated data. 
// Use second function to free it again. 
// Note that some (string) members of ftdi_eeprom are malloc()'ed as well. 
struct ftdi_eeprom *USB_ReadFTDIEEPROM(void);
void USB_FreeFTDIEEPROM(struct ftdi_eeprom *eep);
// Pass eep=NULL to just erase the EEPROM. 
// Returns 1 if EEPROM content is invalid. 
int USB_WriteFTDIEEPROM(struct ftdi_eeprom *eep);


#ifdef __cplusplus
}  // extern "C"
#endif

#endif  /* _USB_ATMEL_PRG_USBIO_H_ */
