/*
 * waveformbuf.h -- USB Atmel programmer: Waveform buffer for USB IO. 
 * 
 * Copyright (c) 2006--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef _USB_ATMEL_PRG_WAVEFORMBUF_H_
#define _USB_ATMEL_PRG_WAVEFORMBUF_H_ 1

#include "usbio.h"
#include <assert.h>
#include <string.h>
#include <inttypes.h>

typedef u_int64_t uint64;


// Waveform buffer stores waveform to be sent to the programmer. 
// This uses mode=1 waveforms with implicit clock. 
class WaveFormBuffer
{
	public:
		// SCK   out TCK
		// MISO  in  TDO
		// MOSI  out TDI
		// Reset out TMS
		// Bit values in mode=1 waveforms: 
		static const uint8 WB_MOSI= 0x01U;
		static const uint8 WB_RESET=0x02U;
		// Corresponding bits for JTAG: 
		static const uint8 WB_TDI=WB_MOSI;
		static const uint8 WB_TMS=WB_RESET;
		
		// Max value for wave_size. 
		static const int WaveBlockMaxSize=256;
		
		// Each wave block corresponds to one mode=1 waveform request 
		// which we will send. 
		struct WaveBlock
		{
			WaveBlock *next;     // Pointer to next block in list. 
			
			int wave_size;       // Wave size: <=256 (state changes / clocks)
			uint8 wave_buf[64];  // Already "compressed" wave data output. 
			uint8 resp_buf[32];  // Read back response buffer. 
			
			uint8 expect_buf[32];  // Expected readback values. 
			uint8 expect_mask[32]; // Expectation mask (0 = don't care). 
			
			// Pointers where to store readback values. 
			// One pointer for each bit or NULL. 
			// The pointers normally point to variables allocated on the 
			// stack of the calling function and should stay valid until 
			// the flush() function is called which will store the readback 
			// values. 
			// NOTE: The local stack wars to which these pointers point should 
			//       be initialized to some value (often zero). 
			// The type of the dest pointers is uint8, uint16, uint32 or 
			// uint64 depending on the top 2 bits of dest_bidx[]. 
			void *dest[256];
			// Lower 6 bits (0..63) are bit index for the readback. 
			// For example, if you want to read in bits 0,1,2 to some var, use 
			//   dest[i]=dest[1+1]=dest[1+2]=&some_var and
			//   dest_bidx[i]=0;  dest_bidx[i+1]=1; dest_bidx[i+2]=2;
			// The top 2 bits represent the type of the dest[] pointers and 
			// values 0,1,2,3 correspond to uint8, uint16, uint32 and uint64. 
			// NOTE: bit index 0 is the least significant bit. 
			uint8 dest_bidx[256];  // Each value: 0..63 and 2 bits for type sel.
			
			// Constructor initializes expect buf and mask to 0 as 
			// well as size, next pointer and dest arrays, . 
			inline WaveBlock();
			inline ~WaveBlock() {}
			
			void *operator new(size_t size);
			void operator delete(void *ptr);
		};
		
	private:
		// List of wave blocks to be sent the next time. 
		// *send_list is to be sent first. 
		WaveBlock *send_list;
		// Current last element of the list. 
		WaveBlock *list_head;
		
		// Number of blocks in send list: 
		int pending_blocks;
		
	public:
		// Statistics: Accumulated number of wave blocks sent: 
		int nblocks_sent;
		// Accumulated size of sent blocks. 
		int totsize_sent;
		
	private:
		// Append passed 2 bits to current wave block. 
		// Expect is 0 to expect nothing; 0x1 to expect a 0 and 0x3 to expect 1.
		// dest and dest_bidx are the corresponding destination values for 
		// the equally named fields in WaveBlock. 
		void _append(uint8 bits,uint8 expect,void *dest,uint8 dest_bidx);
		
		// Remove and delete first entry in the list. 
		void _RemoveFirst();
		
		// Internally used by the flush() function when the WaveBlock's 
		// readback has been received to compare it to the expected values 
		// and to store the result in the dest[] pointers. Returns error 
		// count for expect mismatches. 
		int _DoStoreCompareReadback(WaveBlock *wb_recv,char jverbose);
		
		// Do not use. 
		WaveFormBuffer(const WaveFormBuffer &);
		void operator=(const WaveFormBuffer &);
	public:
		WaveFormBuffer();
		~WaveFormBuffer()
			{  clear(1);  }
		
		// List is empty?
		inline bool IsEmpty() const
			{  return(!list_head);  }
		
		// Number of pending blocks?
		inline int NPending() const
			{  return(pending_blocks);  }
		
		// Clear will remove all elements from the list. 
		void clear(int force_free=0);
		
		// Clock in the passed bits (WB_TDI,WB_TMS). 
		inline void clock(uint8 bits)
			{  _append(bits,0,NULL,0);  }
		// Clock in the passed bits (WB_TDI,WB_TMS) and expect return 
		// value expect=0 or 1. 
		inline void clock(uint8 bits,uint8 expect)
			{  _append(bits,expect ? 0x03U : 0x01U,NULL,0);  }
		// Version of clock(bits) with readback. 
		inline void clock(uint8 bits,uint8  *dest,uint8 bidx)  { _append(bits,0,dest,bidx|(0U<<6));  }
		inline void clock(uint8 bits,uint16 *dest,uint8 bidx)  { _append(bits,0,dest,bidx|(1U<<6));  }
		inline void clock(uint8 bits,uint32 *dest,uint8 bidx)  { _append(bits,0,dest,bidx|(2U<<6));  }
		inline void clock(uint8 bits,uint64 *dest,uint8 bidx)  { _append(bits,0,dest,bidx|(3U<<6));  }
		
		// Actually perform the IO. 
		// Will send as many waveform requests as necessary to get all 
		// the queued data across and will wait for all the responses. 
		// Responses are compared to expected readback values and 
		// errors are written. 
		// Responses are also stored in the destination buffers/bits as 
		// specified when calling clock(). 
		// Returns total error count. 
		// Use jverbose='s' for SVF files. 
		// wave_delay is passed to the waveform IO. 
		// A value of 0 means 1MHz, 2 means 0.5 MHz, 3 -> 0.33 MHz and so on. 
		int flush(uint8 wave_delay,char jverbose='\0');
};

#endif  /* _USB_ATMEL_PRG_WAVEFORMBUF_H_ */
