/*
 * atmel.cc -- USB Atmel programmer Atmel device. 
 * 
 * Copyright (c) 2005--2009 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "config.h"
#include "atmel.h"
#include "buffer.h"
#include <assert.h>
#include <string.h>

#include <sys/time.h>


namespace { static const uint32
//=============================================================================
// List of common Atmel commands
// -----------------------------
	AT_CMD_ProgEn      = 0xAC530000U, // 10101100 01010011 xxxxxxxx xxxxxxxx
	AT_CMD_ChipErase   = 0xAC800000U, // 10101100 100xxxxx xxxxxxxx xxxxxxxx
	AT_CMD_ReadFlashL  = 0x20000000U, // 00100000 aaaaaaaa bbbbbbbb oooooooo
	AT_CMD_ReadFlashH  = 0x28000000U, // 00101000 aaaaaaaa bbbbbbbb oooooooo
	AT_CMD_ReadEEPROM  = 0xA0000000U, // 10100000 aaaaaaaa bbbbbbbb oooooooo
	AT_CMD_WriteFlashL = 0x40000000U, // 01000000 aaaaaaaa bbbbbbbb iiiiiiii
	AT_CMD_WriteFlashH = 0x48000000U, // 01001000 aaaaaaaa bbbbbbbb iiiiiiii
	AT_CMD_WritePage   = 0x4C000000U, // 01001100 aaaaaaaa bbxxxxxx iiiiiiii
	AT_CMD_WriteEEPROM = 0xC0000000U, // 11000000 aaaaaaaa bbbbbbbb iiiiiiii
	AT_CMD_ReadLock    = 0x58000000U, // 01011000 xxxxxxxx xxxxxxxx oooooooo
	AT_CMD_WriteLock   = 0xACE0FF00U, // 10101100 111xxiix xxxxxxxx iiiiiiii
	AT_CMD_ReadFuse    = 0x50000000U, // 01010000 00000000 xxxxxxxx oooooooo
	AT_CMD_ReadFuseH   = 0x58080000U, // 01011000 00001000 xxxxxxxx oooooooo
	AT_CMD_ReadFuseX   = 0x50080000U, // 01010000 00001000 xxxxxxxx oooooooo
	AT_CMD_WriteFuse   = 0xACA0FF00U, // 10101100 10100000 xxxxxxxx iiiiiiii
	AT_CMD_WriteFuseH  = 0xACA8FF00U, // 10101100 10101000 xxxxxxxx iiiiiiii
	AT_CMD_WriteFuseX  = 0xACA4FF00U, // 10101100 10100100 xxxxxxxx iiiiiiii
	AT_CMD_ReadSig     = 0x30000000U, // 00110000 xxxxxxxx xxxxxxbb oooooooo
	AT_CMD_ReadCalib   = 0x38000000U; // 00111000 00xxxxxx 00000000 oooooooo
//=============================================================================
// Fuse and lock selection mask values to be OR'ed together: 
static const int 
   FSM_Lock= 0x01,
   FSM_FuseL=0x02,
   FSM_FuseH=0x04,
   FSM_FuseX=0x08,
   FSM_Calib=0x10;  // for reading only
//=============================================================================
}


static uint8 StrToByte(const char *str,FILE *out,int *errcnt)
{
	int base=-1;
	const char *start=str;
	if(!strncmp(str,"0b",2))
	{  base=2;  start+=2;  }
	else if(!strncmp(str,"0x",2))
	{  base=16;  start+=2;  }
	if(base<0)
	{  fprintf(out,"Invalid spec \"%s\". Use 0x or 0b prefix.\n",str);
		++*errcnt;  return(0);  }
	
	char *end=NULL;
	unsigned long val=strtoul(start,&end,base);
	if(*end)
	{  fprintf(out,"Invalid spec \"%s\".\n",str);
		++*errcnt;  return(0);  }
	if(val>0xff)
	{  fprintf(out,"Value \"%s\" out of range.\n",str);
		++*errcnt;  return(0);  }
	
	return((uint8)(val&0xffU));
}


static void DumpLockHeader(FILE *out,uint8 lock)
{
	fprintf(out,"Lock bits: 0x%02x  (programmed=0)\n",lock);
}

// Dump the basic first 2 lock bits which seem to be always the same. 
static void DumpLock_BasicTwoBits(FILE *out,uint8 lock)
{
	const char *mem_lock_msg=NULL;
	switch(lock & 0x03)
	{
		case 0:  mem_lock_msg="memory read/write disabled";  break;
		case 2:  mem_lock_msg="memory write protected";  break;
		case 3:  mem_lock_msg="unrestricted memory access";  break;
		case 1:  // Value 1 not documented. 
		default:  mem_lock_msg="unknown state";  break;
	}
	
	fprintf(out,
		"  LB     2,1: %d%d   (%s)\n"
		,(lock & 0x02) ? 1 : 0,(lock & 0x01) ? 1 : 0,mem_lock_msg);
}


// Standard lock byte dump function for ATMega type of chips; should work 
// for most current ones. 
static void DumpLock_StdMega(FILE *out,uint8 lock,const char *page_text)
{
	DumpLockHeader(out,lock);
	
	fprintf(out,
		"  BLB1   2,1: %d%d   (%s)\n"
		"  BLB0   2,1: %d%d   (%s)\n"
		,(lock & 0x20) ? 1 : 0,(lock & 0x10) ? 1 : 0,page_text
		,(lock & 0x08) ? 1 : 0,(lock & 0x04) ? 1 : 0,page_text );
	
	DumpLock_BasicTwoBits(out,lock);
}


static void DumpFuseHeader(FILE *out,uint8 fuse)
{
	fprintf(out,"Fuse low bits: 0x%02x  (programmed=0)\n",fuse);
}

// Dump the common CKSEL and SUT values. 
static void DumpFuse_CommonCKSELAndSUT(FILE *out,uint8 fuse)
{
	const char *cksel_msg=NULL;
	switch(fuse & 0x0f)
	{
		case 0:  cksel_msg="external clock source"; break;  // OK
		case 1:  cksel_msg="internal RC osc at 1 MHz";  break;  // OK
		case 2:  cksel_msg="internal RC osc at 2 MHz";  break;  // OK
		case 3:  cksel_msg="internal RC osc at 4 MHz";  break;  // OK
		case 4:  cksel_msg="internal RC osc at 8 MHz";  break;  // OK
		case 5:  cksel_msg="external RC osc 0.1 .. 0.9 MHz";  break;  // OK
		case 6:  cksel_msg="external RC osc 0.9 .. 3.0 MHz";  break;  // OK
		case 7:  cksel_msg="external RC osc 3.0 .. 8.0 MHz";  break;  // OK
		case 8:  cksel_msg="external RC osc 8.0 .. 12 MHz";  break;   // OK
		case 9:  cksel_msg="external low-freq crystal (32 kHz)";  break;  // OK
		case 10: // and...
		case 11: cksel_msg="ext. ceramic osc, 0.4 .. 0.9 MHz";  break;
		case 12: // and...
		case 13: cksel_msg="ext. xtal/ceramic osc, 0.9 .. 3.0 MHz";  break;
		case 14: // and...
		case 15: cksel_msg="ext. xtal/ceramic osc, 3.0 .. 8/16 MHz";  break;
		default:  cksel_msg="invalid";  break;
	}
	
	fprintf(out,
		"  SUT   1,0:  %d%d\n"
		"  CKSEL 3..0: %d%d%d%d (%s)\n"
		,(fuse & 0x20) ? 1 : 0,(fuse & 0x10) ? 1 : 0
		,(fuse & 0x08) ? 1 : 0,(fuse & 0x04) ? 1 : 0,
			(fuse & 0x02) ? 1 : 0,(fuse & 0x01) ? 1 : 0,cksel_msg);
}

// Standard fuse low byte dump function for ATMega type of chips; should work 
// for most current ones. 
static void DumpFuseL_StdMega(FILE *out,uint8 fuse,
	const char *bod0_txt,const char *bod1_txt)
{
	DumpFuseHeader(out,fuse);
	
	fprintf(out,
		"  BODLEVEL:   %d    (%s)\n"
		"  BODEN:      %d    (brown-out detector %s)\n"
		,(fuse & 0x80) ? 1 : 0,      // <-- BODLEVEL
			(fuse & 0x80) ? bod1_txt : bod0_txt
		,(fuse & 0x40) ? 1 : 0,      // <-- BODEN
			(fuse & 0x40) ? "disabled" : "enabled");
	
	DumpFuse_CommonCKSELAndSUT(out,fuse);
}


// Used e.g. by ATMega162. 
static void DumpFuseL_CKDIV8_CKOUT_SUT_CKSEL(FILE *out,uint8 fuse,
	const char *ck_ds_page,const char *cksel_msg3)
{
	fprintf(out,
		"  CKDIV8:     %d    (%s)\n"
		"  CKOUT:      %d    (clock output on PB0 %s)\n"
		,(fuse & 0x80) ? 1 : 0,      // <-- CKDIV8
			(fuse & 0x80) ? "do not divide clock" : "divide clock by 8"
		,(fuse & 0x40) ? 1 : 0,      // <-- CKOUT
			(fuse & 0x40) ? "disabled" : "enabled");
	
	const char *cksel_msg=NULL;  // See p.35
	const char *cksel_msg2="";
	switch(fuse & 0x07)
	{
		case 0:  cksel_msg="external clock source"; break;
		case 1:  cksel_msg="reserved";  break;
		case 2:  cksel_msg="internal RC osc at 8 MHz";  break;
		case 3:  cksel_msg=cksel_msg3;  break;
		default: cksel_msg="please see data sheet ";
		         cksel_msg2=ck_ds_page;  break;
	}
	
	fprintf(out,
		"  SUT   1,0:  %d%d\n"
		"  CKSEL 3..0: %d%d%d%d (%s%s)\n"
		,(fuse & 0x20) ? 1 : 0,(fuse & 0x10) ? 1 : 0
		,(fuse & 0x08) ? 1 : 0,(fuse & 0x04) ? 1 : 0,
			(fuse & 0x02) ? 1 : 0,(fuse & 0x01) ? 1 : 0,
			cksel_msg,cksel_msg2);
}


struct AtmelDevice_ATMega8515 : AtmelDevice
{
	AtmelDevice_ATMega8515(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega8515")
	{
		flash_size=4096;  // words
		eeprom_size=512;  // bytes
		page_size=64;     // bytes (32 words)
		fuse_lock_bytes=3;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.177");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseL_StdMega(out,fuse,
			"brown-out level 4.0V, see p.45","brown-out level 2.7V, see p.45");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03)
		{
			case 0:  bootsz_msg="boot size 1024 bytes, see p.175";  break;
			case 1:  bootsz_msg="boot size 512 bytes, see p.175";  break;
			case 2:  bootsz_msg="boot size 256 bytes, see p.175";  break;
			case 3:  bootsz_msg="boot size 128 bytes, see p.175";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  S8515C:     %d    (AT90S4414/8515 compat mode %s)\n"
			"  WDTON:      %d    (%s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  CKOPT:      %d    (%s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x80) ? "disabled" : "enabled"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "WDT off" : "WDT always on"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "small output swing, see p.34" : 
					"rail-to-rail swing, see p.34"
			,(fuseH & 0x08) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{  return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH));  }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH,
			args,nargs));  }
};


struct AtmelDevice_ATMega162 : AtmelDevice
{
	AtmelDevice_ATMega162(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega162")
	{
		flash_size=8192;   // words
		eeprom_size=512;   // bytes
		page_size=128;     // bytes (64 words)
		fuse_lock_bytes=4;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.230");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		DumpFuseL_CKDIV8_CKOUT_SUT_CKSEL(out,fuse," p.34ff","reserved");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03)
		{
			case 0:  bootsz_msg="boot size 1024 words, see p.228";  break;
			case 1:  bootsz_msg="boot size 512 words, see p.228";  break;
			case 2:  bootsz_msg="boot size 256 words, see p.228";  break;
			case 3:  bootsz_msg="boot size 128 words, see p.228";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  OCDEN:      %d    (OCD %s)\n"
			"  JTAGEN:     %d    (JTAG %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  WDTON:      %d    (%s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x10) ? "disabled" : "enabled [never ship device like that!]"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "WDT off" : "WDT always on"
			,(fuseH & 0x08) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		const char *bod_string=NULL;
		switch((fuseX>>1)&0x07)
		{
			case 0:  // Refer to p.49 (Table 19). 
			case 1:
			case 2:  bod_string="reserved setting!";  break;
			case 3:  bod_string="trigger level: 2.3V (ATMega162V only)";  break;
			case 4:  bod_string="trigger level: 4.3V";  break;
			case 5:  bod_string="trigger level: 2.7V";  break;
			case 6:  bod_string="trigger level: 1.8V (ATMega162V only)";  break;
			case 7:  bod_string="disabled";  break;
		}
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  M161C:         %d    (ATMega161 compatibility mode %sabled)\n"
			"  BODLEVEL 2..0: %d%d%d  (BOD %s)\n"
			,fuseX
			,(fuseX & 0x10) ? 1 : 0,(fuseX & 0x10) ? "dis" : "en"
			,(fuseX & 0x08) ? 1 : 0,(fuseX & 0x04) ? 1 : 0,
				(fuseX & 0x02) ? 1 : 0,bod_string);
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{  return(_DoDumpFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX));  }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,args,nargs));  }
};


struct AtmelDevice_ATMega32 : AtmelDevice
{
	AtmelDevice_ATMega32(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega32")
	{
		flash_size=16384;   // words
		eeprom_size=1024;   // bytes
		page_size=128;      // bytes (64 words)
		fuse_lock_bytes=3;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.256");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseL_StdMega(out,fuse,
			"brown-out level 4.0V, see p.37","brown-out level 2.7V, see p.37");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03U)
		{
			case 0:  bootsz_msg="boot size 4096 bytes, see p.255";  break;
			case 1:  bootsz_msg="boot size 1024 bytes, see p.255";  break;
			case 2:  bootsz_msg="boot size 512 bytes, see p.255";  break;
			case 3:  bootsz_msg="boot size 256 bytes, see p.255";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  OCDEN:      %d    (OCD %s)\n"
			"  JTAGEN:     %d    (JTAG %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  CKOPT:      %d    (see p.26)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x10) ? "disabled" : "enabled [never ship device like that!]"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0    // <-- CKOPT
			,(fuseH & 0x08) ? 1 : 0,   // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{  return(_DoDumpFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH));  }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH,args,nargs));  }
};


#if 0   /* Officially not supported - PEN pin. */
struct AtmelDevice_ATMega64 : AtmelDevice
{
	AtmelDevice_ATMega64(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega64")
	{
		flash_size=32768;   // words
		eeprom_size=2048;   // bytes
		page_size=256;      // bytes (128 words)
		fuse_lock_bytes=4;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.290");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseL_StdMega(out,fuse,
			"brown-out level 4.0V, see p.52","brown-out level 2.7V, see p.52");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03U)
		{
			case 0:  bootsz_msg="boot size 4096 bytes, see p.289";  break;
			case 1:  bootsz_msg="boot size 2048 bytes, see p.289";  break;
			case 2:  bootsz_msg="boot size 1024 bytes, see p.289";  break;
			case 3:  bootsz_msg="boot size 512 bytes, see p.289";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  OCDEN:      %d    (OCD %s)\n"
			"  JTAGEN:     %d    (JTAG %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  CKOPT:      %d    (see p.38ff)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x10) ? "disabled" : "enabled [never ship device like that!]"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0    // <-- CKOPT
			,(fuseH & 0x08) ? 1 : 0,   // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  M103C:      %d    (ATMega103 compatibility mode %sabled)\n"
			"  WDTON:      %d    (watchdog %s)\n"
			,fuseX,
			(fuseX & 0x02) ? 1 : 0,
				(fuseX & 0x02) ? "dis" : "en",
			(fuseX & 0x01) ? 1 : 0,
				(fuseX & 0x01) ? "off" : "always on");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{  return(_DoDumpFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX));  }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,args,nargs));  }
};
#endif


struct AtmelDevice_ATMega325 : AtmelDevice
{
	AtmelDevice_ATMega325(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega325")
	{
		flash_size=16384;   // words
		eeprom_size=1024;   // bytes
		page_size=128;      // bytes (128 words)
		fuse_lock_bytes=4;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.290");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		
		const char *cksel_msg=NULL;
		switch(fuse & 0x0f)
		{
			case 0:  cksel_msg="external clock source"; break;  // OK
			case 1:  cksel_msg="reserved";  break;  // OK
			case 2:  cksel_msg="internal RC osc at 8 MHz";  break;  // OK
			case 3:  // OK
			case 4:  // OK
			case 5:  cksel_msg="reserved";  break;
			case 6:  // OK
			case 7:  cksel_msg="external low-freq crystal (32 kHz)";  break;  // OK
			case 8:  // and...
			case 9:  cksel_msg="ext. ceramic osc, 0.4 .. 0.9 MHz";  break;
			case 10: // and...
			case 11: cksel_msg="ext. xtal/ceramic osc, 0.9 .. 3.0 MHz";  break;
			case 12: // and...
			case 13: cksel_msg="ext. xtal/ceramic osc, 3.0 .. 8 MHz";  break;
			case 14: // and...
			case 15: cksel_msg="ext. xtal/ceramic osc, >8 MHz";  break;
			default:  cksel_msg="invalid";  break;
		}
		
		fprintf(out,
			"  CKDIV8:     %d    (clock division by 8 %s)\n"
			"  CKOUT:      %d    (clock output on CKOUT %s)\n"
			"  SUT   1,0:  %d%d\n"
			"  CKSEL 3..0: %d%d%d%d (%s)\n"
			,(fuse & 0x80) ? 1 : 0,
				(fuse & 0x80) ? "disabled" : "enabled"
			,(fuse & 0x40) ? 1 : 0,
				(fuse & 0x40) ? "disabled" : "enabled"
			,(fuse & 0x20) ? 1 : 0,(fuse & 0x10) ? 1 : 0
			,(fuse & 0x08) ? 1 : 0,(fuse & 0x04) ? 1 : 0,
				(fuse & 0x02) ? 1 : 0,(fuse & 0x01) ? 1 : 0,cksel_msg);
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03U)
		{
			case 0:  bootsz_msg="boot size 2048 bytes, see p.261";  break;
			case 1:  bootsz_msg="boot size 1024 bytes, see p.261";  break;
			case 2:  bootsz_msg="boot size 512 bytes, see p.261";  break;
			case 3:  bootsz_msg="boot size 256 bytes, see p.261";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  OCDEN:      %d    (OCD %s)\n"
			"  JTAGEN:     %d    (JTAG %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  WDTON:      %d    (watchdog %s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x10) ? "disabled" : "enabled [never ship device like that!]"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "off" : "always on"
			,(fuseH & 0x08) ? 1 : 0,   // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		const char *bodlevel_str=NULL;
		switch((fuseX>>1) & 0x03U)
		{
			case 0:  bodlevel_str="level 4.3 V";  break;
			case 1:  bodlevel_str="level 2.7 V";  break;
			case 2:  bodlevel_str="level 1.8 V";  break;
			case 3:  bodlevel_str="disabled";  break;
			default:  bodlevel_str="invalid";  break;
		}
		
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  BODLVL 1,0: %d%d   (BOD %s)\n"
			"  RSTDISBL:   %d    (%s)  [leave 1=unprogrammed!, p.265]\n"
			,fuseX
			,(fuseX & 0x04) ? 1 : 0,(fuseX & 0x02) ? 1 : 0,bodlevel_str
			,(fuseX & 0x01) ? 1 : 0,
				(fuseX & 0x80) ? "RESET pin enabled" : "RESET is normal IO pin");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{  return(_DoDumpFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX));  }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,args,nargs));  }
};


struct AtmelDevice_AT90S2313 : AtmelDevice
{
	AtmelDevice_AT90S2313(const char *_vendor) : 
		AtmelDevice(_vendor,"AT90S2313")
	{
		flash_size=1024;  // words
		eeprom_size=128;  // bytes
		page_size=0;      // bytes (no paging)
		fuse_lock_bytes=1;
		has_erase_busy=0;
	}
	
	int DumpFuseLock(FILE *out)
	{
		// The AT90S2313 does not allow to read fuse and lock bits via SPI. 
		fprintf(out,"AT90S2313: Fuse and lock bits not readable via SPI.\n");
		return(1);
	}
};


struct AtmelDevice_ATTiny2313 : AtmelDevice
{
	AtmelDevice_ATTiny2313(const char *_vendor) : 
		AtmelDevice(_vendor,"ATTiny2313")
	{
		flash_size=1024;  // words
		eeprom_size=128;  // bytes
		page_size=32;     // bytes
		fuse_lock_bytes=4;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
	{
		DumpLockHeader(out,lock);
		DumpLock_BasicTwoBits(out,lock);
	}
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		
		fprintf(out,
			"  CKDIV8:     %d    (clock division by 8 %s)\n"
			"  CKOUT:      %d    (clock output on CKOUT %s)\n"
			,(fuse & 0x80) ? 1 : 0,
				(fuse & 0x80) ? "disabled" : "enabled"
			,(fuse & 0x40) ? 1 : 0,
				(fuse & 0x40) ? "disabled" : "enabled");
		
		DumpFuse_CommonCKSELAndSUT(out,fuse);
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bodlvl_msg=NULL;
		switch((fuseH>>1) & 0x07U)
		{
			case 7:  bodlvl_msg="brown-out detection disabled";  break;
			case 6:  bodlvl_msg="brown-out detection level at 1.8V";  break;
			case 5:  bodlvl_msg="brown-out detection level at 2.7V";  break;
			case 4:  bodlvl_msg="brown-out detection level at 4.3V";  break;
			case 3:  // fall..
			case 2:  //   ...
			case 1:  //  ...through
			case 0:  bodlvl_msg="brown-out detection: reserved value";  break;
			default:  bodlvl_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  DWEN:       %d    (debugWire %s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  WDTON:      %d    (%s)\n"
			"  BODLVL 210: %d%d%d  (%s)\n"
			"  RSTDISBL:   %d    (external reset %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x80) ? "disabled" : "enabled"
			,(fuseH & 0x40) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x40) ? "erased" : "preserved"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "WDT off" : "WDT always on"
			,(fuseH & 0x08) ? 1 : 0,(fuseH & 0x04) ? 1 : 0,
				(fuseH & 0x02) ? 1 : 0,
				bodlvl_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "enabled" : "disabled");
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		// Unfortunately, the data sheet does not explicitly explain 
		// if a logic 0 in SELFPRGEN means "enabled" or "disabled". 
		// Default is 1 (unprogrammed). 
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  SELFPRGEN:  %d    (self-programming [1=disabled??])\n"
			,fuseX,
			(fuseX & 0x01) ? 1 : 0);
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
	{ return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX)); }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,args,nargs));  }
};


struct AtmelDevice_ATTiny24_44_84 : AtmelDevice
{
	AtmelDevice_ATTiny24_44_84(const char *_vendor,char which) : 
		AtmelDevice(_vendor,
			which=='2' ? "ATTiny24" : which=='4' ? "ATTiny44" : 
				which=='8' ? "ATTiny84" : "<unknown>")
	{
		switch(which)
		{
			case '2':  // ATTiny24
				flash_size=1024;  // words
				eeprom_size=128;  // bytes
				page_size=32;     // bytes
				break;
			case '4':  // ATTiny44
				flash_size=2048;  // words
				eeprom_size=256;  // bytes
				page_size=64;     // bytes
				break;
			case '8':  // ATTiny44
				flash_size=4096;  // words
				eeprom_size=512;  // bytes
				page_size=64;     // bytes
				break;
			default:  assert(0);
		}
		
		fuse_lock_bytes=4;  // lock, fuse low, high, extended
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
	{
		DumpLockHeader(out,lock);
		DumpLock_BasicTwoBits(out,lock);
	}
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		DumpFuseL_CKDIV8_CKOUT_SUT_CKSEL(out,fuse,"","reserved");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bodlvl_msg=NULL;
		switch(fuseH & 0x07U)
		{
			case 7:  bodlvl_msg="brown-out detection disabled";  break;
			case 6:  bodlvl_msg="brown-out detection level at 1.8V";  break;
			case 5:  bodlvl_msg="brown-out detection level at 2.7V";  break;
			case 4:  bodlvl_msg="brown-out detection level at 4.3V";  break;
			case 3:  bodlvl_msg="brown-out detection level at 2.3V";  break;
			case 2:  bodlvl_msg="brown-out detection level at 2.2V";  break;
			case 1:  bodlvl_msg="brown-out detection level at 1.9V";  break;
			case 0:  bodlvl_msg="brown-out detection level at 2.0V";  break;
			default:  bodlvl_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  RSTDISBL:   %d    (external reset %s)\n"
			"  DWEN:       %d    (debugWire %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  WDTON:      %d    (%s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BODLVL 210: %d%d%d  (%s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x80) ? "enabled" : "disabled"  // "ext reset disable"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"  // "DW enable"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "WDT off" : "WDT always on"
			,(fuseH & 0x08) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,
				(fuseH & 0x01) ? 1 : 0,
				bodlvl_msg);
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  SELFPRGEN:  %d    (self-programming %sabled)\n"
			,fuseX
			,(fuseX & 0x01) ? 1 : 0,(fuseX & 0x10) ? "dis" : "en");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
	{ return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX)); }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,args,nargs));  }
};


struct AtmelDevice_ATTiny13 : AtmelDevice
{
	AtmelDevice_ATTiny13(const char *_vendor) : 
		AtmelDevice(_vendor,"ATTiny13")
	{
		flash_size=512;   // words
		eeprom_size=64;   // bytes
		page_size=32;     // bytes
		fuse_lock_bytes=3;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
	{
		DumpLockHeader(out,lock);
		DumpLock_BasicTwoBits(out,lock);
	}
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		
		const char *cksel_msg=NULL;  // See p.35
		switch(fuse & 0x03U)
		{
			case 0:  cksel_msg="external clock source"; break;
			case 1:  cksel_msg="internal RC osc at 4.8 MHz";  break;
			case 2:  cksel_msg="internal RC osc at 9.6 MHz";  break;
			case 3:  cksel_msg="internal RC osc at 128 kHz";  break;
		}
		
		fprintf(out,
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  WDTON:      %d    (watchdog %s)\n"
			"  CKDIV8:     %d    (clock division by 8 %s)\n"
			"  SUT   1,0:  %d%d\n"
			"  CKSEL 1,0:  %d%d   (%s)\n"
			,(fuse & 0x80) ? 1 : 0,
				(fuse & 0x80) ? "disabled" : "enabled"
			,(fuse & 0x40) ? 1 : 0,
				(fuse & 0x40) ? "erased" : "preserved"
			,(fuse & 0x20) ? 1 : 0,
				(fuse & 0x20) ? "off" : "always on"
			,(fuse & 0x10) ? 1 : 0,
				(fuse & 0x10) ? "disabled" : "enabled"
			,(fuse & 0x08) ? 1 : 0,(fuse & 0x04) ? 1 : 0
			,(fuse & 0x02) ? 1 : 0,(fuse & 0x01) ? 1 : 0,cksel_msg);
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bod_msg=NULL;
		switch((fuseH>>1) & 0x03U)
		{
			case 0:  bod_msg="level at 1.8V";  break;
			case 1:  bod_msg="level at 2.7V";  break;
			case 2:  bod_msg="level at 4.3V";  break;
			case 3:  bod_msg="disabled";       break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  SELFPRGEN:  %d    (self programming %sabled)\n"
			"  DWEN:       %d    (debug wire %sabled)\n"
			"  BODLVL1,0:  %d%d   (brown-out detector %s)\n"
			"  RSTDISBL:   %d    (reset %sabled)\n"
			,fuseH
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "dis" : "en"
			,(fuseH & 0x08) ? 1 : 0,
				(fuseH & 0x08) ? "dis" : "en"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,
				bod_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "en" : "dis");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
	{ return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH)); }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH,args,nargs));  }
};


struct AtmelDevice_ATTiny26 : AtmelDevice
{
	AtmelDevice_ATTiny26(const char *_vendor) : 
		AtmelDevice(_vendor,"ATTiny26")
	{
		flash_size=1024;  // words
		eeprom_size=128;  // bytes
		page_size=32;     // bytes
		fuse_lock_bytes=3;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
	{
		DumpLockHeader(out,lock);
		DumpLock_BasicTwoBits(out,lock);
	}
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		
		fprintf(out,
			"  PLLCK:      %d    (PLL for internal clock %s)\n"
			"  CKOPT:      %d    (osc option %s; see docu)\n"
			,(fuse & 0x80) ? 1 : 0,
				(fuse & 0x80) ? "disabled" : "enabled"
			,(fuse & 0x40) ? 1 : 0,
				(fuse & 0x40) ? "disabled" : "enabled");
		
		DumpFuse_CommonCKSELAndSUT(out,fuse);
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  RSTDISBL:   %d    (external reset %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BODLVL:     %d    (brown-out detector level %s)\n"
			"  BODEN:      %d    (brown-out detector %s)\n"
			,fuseH
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "enabled" : "disabled"
			,(fuseH & 0x08) ? 1 : 0,
				(fuseH & 0x08) ? "disabled" : "enabled"
			,(fuseH & 0x04) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x04) ? "erased" : "preserved"
			,(fuseH & 0x02) ? 1 : 0,
				(fuseH & 0x02) ? "2.7V" : "4.0V"
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "disabled" : "enabled");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
	{ return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH)); }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,
			FSM_Lock|FSM_FuseL|FSM_FuseH,args,nargs));  }
};


struct AtmelDevice_ATMega8 : AtmelDevice
{
	AtmelDevice_ATMega8(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega8")
	{
		flash_size=4096;  // words
		eeprom_size=512;  // bytes
		page_size=64;     // bytes (32 words)
		fuse_lock_bytes=3;
		has_erase_busy=1;
		calibration_bytes=4;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.219");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseL_StdMega(out,fuse,
			"brown-out level 4.0V, see p.36","brown-out level 2.6V, see p.36");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03)
		{
			case 0:  bootsz_msg="boot size 1024 words, see p.217";  break;
			case 1:  bootsz_msg="boot size 512 words, see p.217";  break;
			case 2:  bootsz_msg="boot size 256 words, see p.217";  break;
			case 3:  bootsz_msg="boot size 128 words, see p.217";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  RSTDISBL:   %d    (%s)  [leave 1=unprogrammed!, p.220]\n"
			"  WDTON:      %d    (%s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  CKOPT:      %d    (%s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x80) ? "RESET pin enabled" : "RESET is normal IO pin"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "WDT off" : "WDT always on"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "small output swing, see p.25" : 
					"rail-to-rail swing, see p.25"
			,(fuseH & 0x08) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{  return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_Calib));  }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH,
			args,nargs));  }
};


struct AtmelDevice_ATMega88 : AtmelDevice
{
	AtmelDevice_ATMega88(const char *_vendor) : 
		AtmelDevice(_vendor,"ATMega88")
	{
		flash_size=4096;  // words
		eeprom_size=512;  // bytes
		page_size=64;     // bytes (32 words)
		fuse_lock_bytes=3;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.177");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		DumpFuseL_CKDIV8_CKOUT_SUT_CKSEL(out,fuse," p.29ff",
			"internal RC osc at 128 kHz");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bodlevel_str=NULL;
		switch(fuseH & 0x07U)
		{
			case 0:  // all
			case 1:  //   these
			case 2:  //      are reserved
			case 3:  bodlevel_str="reserved (see p.308)";  break;
			case 4:  bodlevel_str="level 4.1-4.5 V";  break;
			case 5:  bodlevel_str="level 2.5-2.9 V";  break;
			case 6:  bodlevel_str="level 1.7-2.0 V";  break;
			case 7:  bodlevel_str="disabled";  break;
			default:  bodlevel_str="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  RSTDISBL:   %d    (%s)  [leave 1=unprogrammed!, p.220]\n"
			"  DWEN:       %d    (debugWire %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  WDTON:      %d    (%s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BODLEVEL 2..0: %d%d%d  (BOD %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x80) ? "RESET pin enabled" : "RESET is normal IO pin"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "WDT off" : "WDT always on"
			,(fuseH & 0x08) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,
				(fuseH & 0x01) ? 1 : 0,bodlevel_str);
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		const char *bootsz_msg=NULL;
		switch((fuseX>>1) & 0x03)
		{
			case 0:  bootsz_msg="boot size 1024 bytes, see p.282";  break;
			case 1:  bootsz_msg="boot size 512 bytes, see p.282";  break;
			case 2:  bootsz_msg="boot size 256 bytes, see p.282";  break;
			case 3:  bootsz_msg="boot size 128 bytes, see p.282";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseX
			,(fuseX & 0x04) ? 1 : 0,(fuseX & 0x02) ? 1 : 0,bootsz_msg
			,(fuseX & 0x01) ? 1 : 0,
				(fuseX & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{ return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX)); }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,
			args,nargs));  }
};


struct AtmelDevice_ATMegaXX4 : AtmelDevice
{
	AtmelDevice_ATMegaXX4(const char *_vendor,int id,const char *name) : 
		AtmelDevice(_vendor,name)
	{
		switch(id)
		{
			case 6440:
			case 6441:
				flash_size=32768;  // words
				eeprom_size=2048;  // bytes
				page_size=256;     // bytes (128 words)
				break;
			case 1641:
				flash_size=8192;   // words
				eeprom_size=512;   // bytes
				page_size=128;     // bytes (128 words)
				break;
			case 3241:
				flash_size=16384;  // words
				eeprom_size=1024;  // bytes
				page_size=128;     // bytes (128 words)
				break;
		}
		fuse_lock_bytes=4;
		has_erase_busy=1;
	}
	
	// Overriding virtual. 
	void _DumpLock(FILE *out,uint8 lock)
		{  DumpLock_StdMega(out,lock,"see p.284");  }
	
	// Overriding virtual. 
	void _DumpFuseL(FILE *out,uint8 fuse)
	{
		DumpFuseHeader(out,fuse);
		DumpFuseL_CKDIV8_CKOUT_SUT_CKSEL(out,fuse," p.28ff",
			"internal RC osc at 128 kHz");
	}
	
	// Overriding virtual. 
	void _DumpFuseH(FILE *out,uint8 fuseH)
	{
		const char *bootsz_msg=NULL;
		switch((fuseH>>1) & 0x03)
		{
			case 0:  bootsz_msg="boot size 4096 bytes, see p.280";  break;
			case 1:  bootsz_msg="boot size 2048 bytes, see p.280";  break;
			case 2:  bootsz_msg="boot size 1024 bytes, see p.280";  break;
			case 3:  bootsz_msg="boot size 512 bytes, see p.280";  break;
			default:  bootsz_msg="invalid";  break;
		}
		
		fprintf(out,
			"Fuse high bits: 0x%02x  (programmed=0)\n"
			"  OCDEN:      %d    (OCD %s)\n"
			"  JTAGEN:     %d    (JTAG %s)\n"
			"  SPIEN:      %d    (SPI %s)  [readonly]\n"
			"  WDTON:      %d    (%s)\n"
			"  EESAVE:     %d    (EEPROM %s at chip erase)\n"
			"  BOOTSZ 1,0: %d%d   (%s)\n"
			"  BOOTRST:    %d    (reset vector at %s)\n"
			,fuseH
			,(fuseH & 0x80) ? 1 : 0,
				(fuseH & 0x10) ? "disabled" : "enabled [never ship device like that!]"
			,(fuseH & 0x40) ? 1 : 0,
				(fuseH & 0x40) ? "disabled" : "enabled"
			,(fuseH & 0x20) ? 1 : 0,
				(fuseH & 0x20) ? "disabled" : "enabled"
			,(fuseH & 0x10) ? 1 : 0,
				(fuseH & 0x10) ? "WDT off" : "WDT always on"
			,(fuseH & 0x08) ? 1 : 0,    // <-- EESAVE
				(fuseH & 0x08) ? "erased" : "preserved"
			,(fuseH & 0x04) ? 1 : 0,(fuseH & 0x02) ? 1 : 0,bootsz_msg
			,(fuseH & 0x01) ? 1 : 0,
				(fuseH & 0x01) ? "address 0" : "boot flash section");
	}
	
	// Overriding virtual. 
	void _DumpFuseX(FILE *out,uint8 fuseX)
	{
		const char *bodlvl_msg=NULL;
		switch(fuseX & 0x07U)
		{
			case 7:  bodlvl_msg="BOD disabled";  break;
			case 6:  bodlvl_msg="BOD trigger level 1.8V";  break;
			case 5:  bodlvl_msg="BOD trigger level 2.7V";  break;
			case 4:  bodlvl_msg="BOD trigger level 4.3V";  break;
			// 3,2,1,0: 
			default:  bodlvl_msg="reserved";  break;
		}
		
		fprintf(out,
			"Fuse extended bits: 0x%02x  (programmed=0)\n"
			"  BODLVL210:  %d%d%d  (%s)\n"
			,fuseX
			,(fuseX & 0x04) ? 1 : 0,(fuseX & 0x02) ? 1 : 0,
				(fuseX & 0x01) ? 1 : 0,bodlvl_msg);
	}
	
	// Overriding virtual. 
	int DumpFuseLock(FILE *out)
		{ return(_DoDumpFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX)); }
	
	// Overriding virtual. 
	int WriteFuseLock(FILE *out,char *const *args,int nargs)
		{  return(_DoWriteFuseLock(out,FSM_Lock|FSM_FuseL|FSM_FuseH|FSM_FuseX,
			args,nargs));  }
};


void AtmelDevice::DumpSupportedDevices(FILE *out)
{
	fprintf(out,
		"Supported Atmel microcontrollers:\n"
		"  AT90S2313, ATTiny2313, ATTiny26, ATMega8, ATMega88, ATMega8515,\n"
		"  ATMega162, ATMega32, ATTiny13, ATTiny24, ATTiny44, ATTiny84,\n"
		"  ATMega644, ATMega164P, ATMega324P, ATMega644P\n");
}


AtmelDevice *GetDeviceBySignature(USBAtmelSignature *sig)
{
	const char *vendor=NULL;
	switch(sig->vendor)
	{
		case 0x1e:  vendor="Atmel";  break;
		default:    vendor="[unknown]";  break;
	}
	
	AtmelDevice *d=NULL;
	switch((int(sig->flash_size)<<8) | int(sig->device))
	{
		case 0x9007:  d=new AtmelDevice_ATTiny13(vendor);    break;
		case 0x9101:  d=new AtmelDevice_AT90S2313(vendor);   break;
		case 0x910a:  d=new AtmelDevice_ATTiny2313(vendor);  break;
		case 0x910b:  d=new AtmelDevice_ATTiny24_44_84(vendor,'2');  break;
		case 0x9109:  d=new AtmelDevice_ATTiny26(vendor);    break;
		case 0x9207:  d=new AtmelDevice_ATTiny24_44_84(vendor,'4');  break;
		case 0x9306:  d=new AtmelDevice_ATMega8515(vendor);  break;
		case 0x930a:  d=new AtmelDevice_ATMega88(vendor);    break;
		case 0x9307:  d=new AtmelDevice_ATMega8(vendor);     break;
		case 0x930C:  d=new AtmelDevice_ATTiny24_44_84(vendor,'8');  break;
		case 0x9404:  d=new AtmelDevice_ATMega162(vendor);   break;
		case 0x9502:  d=new AtmelDevice_ATMega32(vendor);    break;
		case 0x9505:  d=new AtmelDevice_ATMega325(vendor);   break;
#if 0   /* Officially not supported - PEN pin */
		case 0x9602:  d=new AtmelDevice_ATMega64(vendor);    break;
#endif
		case 0x9609:  d=new AtmelDevice_ATMegaXX4(vendor,6440,"ATMega644");  break;
		case 0x940A:  d=new AtmelDevice_ATMegaXX4(vendor,1641,"ATMega164P"); break;
		case 0x9508:  d=new AtmelDevice_ATMegaXX4(vendor,3241,"ATMega324P"); break;
		case 0x960A:  d=new AtmelDevice_ATMegaXX4(vendor,6441,"ATMega644P"); break;
#if 0
		case 0x9203:
			flash_size=2048;
			device="AT90S4433";
			eeprom_size=256;
			config->device=AD_AT90S4433;
			config->page_size=0;
			config->fuse_lock_bytes=2;
			config->has_erase_busy=0;
			break;
		case 0x9401:
			device="ATMega161";
			eeprom_size=512;
			config->device=AD_ATMega161;
			config->page_size=128;
			config->fuse_lock_bytes=2;
			config->has_erase_busy=1;
			break;
#endif
		// default: d stays NULL;
	}
	
	return(d);
}


void *AtmelDevice::operator new(size_t size)
{
	void *ptr=malloc(size);
	assert(ptr);
	return(ptr);
}

void AtmelDevice::operator delete(void *ptr)
{
	if(ptr)  free(ptr);
}


void AtmelDevice::DumpIdentification(FILE *out)
{
	char psize[24];
	if(page_size)
	{  snprintf(psize,24,"%u bytes",page_size); }
	else
	{  strcpy(psize,"[no pages]");  }
	
	fprintf(out,
		"Atmel chip:\n"
		"  Name: %s  (Vendor: %s)\n"
		"  Flash:  %6u bytes  [0x%x words]   Pages: %s\n"
		"  EEPROM: %6u bytes\n"
		,name,vendor,
		flash_size*2,flash_size,psize,
		eeprom_size );
}


int AtmelDevice::EraseAll()
{
	printf("Erasing atmel memories... ");
	fflush(stdout);
	
	uint8 status; // We can ignore this...
	USB_Status rv=USB_AtmelExecSingleCommand(AT_CMD_ChipErase,&status);
	if(rv)
	{
		printf("Failed\n");
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
		return(1);
	}
	
	// Anyways, we need to make a little break. 
	// This is important, otherwise queued commands like 
	// "erase ; writeflash firmware.bin" will actually fail!
	usleep(20000);
	
	// It seems some devices (like the 90S2313 and also the tiny2313) are in 
	// strange states after erasing...
	// At least we cannot simply go on and write the flash here. 
	// Chips which have erase busy may expect to see a flash read to poll 
	// for completion. Those who dont't may need a re-connect. 
	if(has_erase_busy)
	{
		printf("OK; check: ");
		
		// We read address 0 and verify that it is 0xff. 
		rv=USB_AtmelExecSingleCommand(AT_CMD_ReadFlashL,&status);
		if(rv)
		{
			printf("Failed\n");
			fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
			return(1);
		}
		if(status!=0xffU)
		{
			printf("Failed\n");
			fprintf(stderr,
				"Error: Read at flash address 0 a value of 0x%x instead of 0xff.\n"
				"Seems erasing has failed. Please verify device.\n"
				,status);
			return(2);
		}
		else
		{  printf("OK\n");  }
	}
	else
	{
		printf("OK\n");
		return(-1);   /* disconnect to force re-connect */
	}
	
	return(0);
}


int AtmelDevice::ReadFlash(DataBuffer *db,int offset,int nwords)
{
	if(offset<0)
	{
		fprintf(stderr,"Offset must be >=0. Corrected to 0.\n");
		offset=0;
	}
	
	db->Alloc(nwords*2);
	db->size=0;  // Nothing read yet. 
	
	printf("Reading flash at 0x%x into buffer \"%s\" (%d bytes) ",
		offset,db->name,nwords*2);  fflush(stdout);
	
	timeval start_tv;
	gettimeofday(&start_tv,NULL);
	USB_Status rv=USB_ReadFlash(/*adr=*/offset,/*size=*/nwords,
		/*dest=*/(uint16*)db->buf);
	if(rv)
	{
		printf("Failed\n");
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
	}
	else
	{
		long msec=msec_elapsed(&start_tv,NULL);
		printf("OK (%ldms, %.1fkb/s)\n",
			msec,double(nwords*2)/double(msec)/1024.0*1000.0);
		
		db->size=nwords*2;
	}
	
	return(rv ? 1 : 0);
}


int AtmelDevice::WriteFlash(const DataBuffer *db,int offset)
{
	if(offset<0)
	{
		fprintf(stderr,"Offset must be >=0. Corrected to 0.\n");
		offset=0;
	}
	
	printf("Writing flash from buffer \"%s\" (%u bytes) at 0x%x ",
		db->name,db->size,offset);  fflush(stdout);
	
	if((size_t)offset+db->size>size_t(flash_size*2))
	{  fprintf(stderr,"Buffer extends beyond flash size (%d/%d bytes). "
		"Expect trouble.\n",offset+db->size,flash_size*2);  }
	
	timeval start_tv;
	gettimeofday(&start_tv,NULL);
	USB_Status rv=USB_WriteFlash(/*adr=*/offset,/*size=*/db->size/2,
		/*src=*/(const uint16*)db->buf,page_size/2);
	if(rv)
	{
		printf("Failed\n");
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
	}
	else
	{
		long msec=msec_elapsed(&start_tv,NULL);
		printf("OK (%ldms, %.1fkb/s)\n",
			msec,double(db->size)/double(msec)/1024.0*1000.0);
	}
	
	return(rv ? 1 : 0);
}


int AtmelDevice::ReadEEPROM(DataBuffer *db,int offset,int nbytes)
{
	if(offset<0)
	{
		fprintf(stderr,"Offset must be >=0. Corrected to 0.\n");
		offset=0;
	}
	
	db->Alloc(nbytes);
	db->size=0;  // Nothing read yet. 
	
	printf("Reading EEPROM at 0x%x into buffer \"%s\" (%d bytes) ",
		offset,db->name,nbytes);  fflush(stdout);
	
	USB_Status rv=USB_ReadEEPROM(/*adr=*/offset,/*size=*/nbytes,
		/*dest=*/(uint8*)db->buf);
	if(rv)
	{
		printf("Failed\n");
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
	}
	else
	{
		printf("OK  \n");
		db->size=nbytes;
	}
	
	return(rv ? 1 : 0);
}


int AtmelDevice::WriteEEPROM(const DataBuffer *db,int offset)
{
	if(offset<0)
	{
		fprintf(stderr,"Offset must be >=0. Corrected to 0.\n");
		offset=0;
	}
	
	printf("Writing EEPROM from buffer \"%s\" (%u bytes) at 0x%x ",
		db->name,db->size,offset);  fflush(stdout);
	
	if((size_t)offset+db->size>size_t(eeprom_size))
	{  fprintf(stderr,"Buffer extends beyond EEPROM size (%d/%d bytes). "
		"Expect trouble.\n",offset+db->size,eeprom_size);  }
	
	USB_Status rv=USB_WriteEEPROM(/*adr=*/offset,/*size=*/db->size,
		/*src=*/(const uint8*)db->buf);
	if(rv)
	{
		printf("Failed\n");
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
	}
	else
	{  printf("OK  \n");  }
	
	return(rv ? 1 : 0);
}


int AtmelDevice::EraseEEPROM()
{
	printf("Writing 0xff into EEPROM (%u bytes) ",eeprom_size);  fflush(stdout);
	
	char *buf=(char*)malloc(eeprom_size);
	assert(buf);
	memset(buf,0xffU,eeprom_size);
	
	USB_Status rv=USB_WriteEEPROM(/*adr=*/0,/*size=*/eeprom_size,
		/*src=*/(const uint8*)buf);
	if(rv)
	{
		printf("Failed\n");
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
	}
	else
	{  printf("OK  \n");  }
	
	free(buf);
	return(rv ? 1 : 0);
}


// Standard query function for fuse low+high+extended and lock, depending 
// on the fuse_select_mask. See FSM_*. 
// Calls virtual _DumpLock(), _DumpFuseL(), _DumpFuseH(), _DumpFuseX() 
// for user output. 
int AtmelDevice::_DoDumpFuseLock(FILE *out,int fuse_select_mask)
{
	uint8 lock,fuse,fuseH,fuseX;
	
	USB_Status rv=USB_OK;
	do {
		if(fuse_select_mask & FSM_Lock)
		{  rv=USB_AtmelExecSingleCommand(AT_CMD_ReadLock,&lock);  }
		if(rv) break;
		if(fuse_select_mask & FSM_FuseL)
		{  rv=USB_AtmelExecSingleCommand(AT_CMD_ReadFuse,&fuse);  }
		if(rv) break;
		if(fuse_select_mask & FSM_FuseH)
		{  rv=USB_AtmelExecSingleCommand(AT_CMD_ReadFuseH,&fuseH);  }
		if(rv) break;
		if(fuse_select_mask & FSM_FuseX)
		{  rv=USB_AtmelExecSingleCommand(AT_CMD_ReadFuseX,&fuseX);  }
	} while(0);
	if(rv)
	{
		fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
		return(1);
	}
	
	// Tell the user; this calls virtual functions...
	if(fuse_select_mask & FSM_Lock)   _DumpLock(out,lock);
	if(fuse_select_mask & FSM_FuseL)  _DumpFuseL(out,fuse);
	if(fuse_select_mask & FSM_FuseH)  _DumpFuseH(out,fuseH);
	if(fuse_select_mask & FSM_FuseX)  _DumpFuseX(out,fuseX);
	
	// Now, read the calibration bytes. 
	if(fuse_select_mask & FSM_Calib)
	{
		uint8 calib[calibration_bytes];
		for(int i=0; i<calibration_bytes; i++)
		{
			rv=USB_AtmelExecSingleCommand(AT_CMD_ReadCalib | (uint32(i)<<8),
				&calib[i]);
			if(rv) break;
		}
		if(rv)
		{
			fprintf(stderr,"Error: %s\n",USB_Status_String(rv));
			return(1);
		}
		
		_DumpCalib(out,calibration_bytes,calib);
	}
	
	return(0);
}


// Standard write function for fuse low+high+extended and lock, 
// depending on the fuse_select_mask. See FSM_*. 
// Calls virtual _DumpLock(), _DumpFuseL(), _DumpFuseH() for user output. 
int AtmelDevice::_DoWriteFuseLock(FILE *out,
	int fuse_select_mask,
	char *const *args,int nargs)
{
	if(!nargs)
	{  fprintf(out,"Try \"setfl help\".\n");  return(0);  }
	
	bool dry_run=1;
	uint8 fuse=0,fuseH=0,fuseX=0,lock=0;
	bool fuse_spec=0,fuseH_spec=0,fuseX_spec=0,lock_spec=0;
	int errors=0;
	for(int i=0; i<nargs; i++)
	{
		if(!strcmp(args[i],"help"))
		{
			fprintf(out,
				"Usage: setfl%s%s%s%s [!]\n",
				(fuse_select_mask & FSM_FuseL) ? " [fuseL=AA]" : "",
				(fuse_select_mask & FSM_FuseH) ? " [fuseH=BB]" : "",
				(fuse_select_mask & FSM_FuseX) ? " [fuseX=CC]" : "",
				(fuse_select_mask & FSM_Lock) ? " [lock=LL]" : "");
			if(fuse_select_mask & FSM_FuseL)
			{ fprintf(out,"  AA:     fuse low byte value (0bNN, 0xNN)\n");  }
			if(fuse_select_mask & FSM_FuseH)
			{ fprintf(out,"  BB:     fuse high byte value (0bNN, 0xNN)\n");  }
			if(fuse_select_mask & FSM_FuseX)
			{ fprintf(out,"  CC:     fuse extended byte value (0bNN, 0xNN)\n");}
			if(fuse_select_mask & FSM_Lock)
			{ fprintf(out,"  LL:     lock byte value (0bNN, 0xNN)\n");  }
			fprintf(out,
				"  !:      if not specified, only print result but do not actually write\n"
				"Only those bytes specified will be modified.\n"
				"NOTE: Values prefixed with 0b binary, 0x in hex, otherwise invalid!\n"
				"      0bNNN values series of 0/1 as dumped by getfl and read in normal\n"
				"      textbook manner (i.e. lines downwards, left-to-right).\n"
				"It is recommended to fill unused fuse values with logical 1.\n");
			return(0);
		}
		else if(!strcmp(args[i],"!"))
		{  dry_run=0;  }
		else if(!strncmp(args[i],"fuseL=",6) && (fuse_select_mask & FSM_FuseL))
		{  fuse=StrToByte(args[i]+6,out,&errors);  fuse_spec=1;  }
		else if(!strncmp(args[i],"fuseH=",6) && (fuse_select_mask & FSM_FuseH))
		{  fuseH=StrToByte(args[i]+6,out,&errors);  fuseH_spec=1;  }
		else if(!strncmp(args[i],"fuseX=",6) && (fuse_select_mask & FSM_FuseX))
		{  fuseX=StrToByte(args[i]+6,out,&errors);  fuseX_spec=1;  }
		else if(!strncmp(args[i],"lock=",5) && (fuse_select_mask & FSM_Lock))
		{  lock=StrToByte(args[i]+5,out,&errors);  lock_spec=1;  }
		else
		{
			fprintf(out,
				"Illegal argument \"%s\". Try \"setfl help\".\n",
				args[i]);
			++errors;
		}
	}
	if(errors)  return(1);
	
	if(lock_spec)   _DumpLock(out,lock);
	if(fuse_spec)   _DumpFuseL(out,fuse);
	if(fuseH_spec)  _DumpFuseH(out,fuseH);
	if(fuseX_spec)  _DumpFuseX(out,fuseX);
	
	if(dry_run)
	{ fprintf(out,"Not actually writing anything (\"!\" not specified).\n"); }
	else if(lock_spec || fuse_spec || fuseH_spec || fuseX_spec)
	{
		fprintf(out,"Writing:");
		
		USB_Status rv=USB_OK;
		do {
			// NOTE: Program fuse bits before lock bits! (p.179). 
			if(fuse_spec)
			{
				fprintf(out," [fuseL]");  fflush(out);
				rv=USB_AtmelExecSingleCommand(
					AT_CMD_WriteFuse | uint32(fuse),NULL);
				if(rv) break;
			}
			if(fuseH_spec)
			{
				fprintf(out," [fuseH]");  fflush(out);
				rv=USB_AtmelExecSingleCommand(
					AT_CMD_WriteFuseH | uint32(fuseH),NULL);
				if(rv) break;
			}
			if(fuseX_spec)
			{
				fprintf(out," [fuseX]");  fflush(out);
				rv=USB_AtmelExecSingleCommand(
					AT_CMD_WriteFuseX | uint32(fuseX),NULL);
				if(rv) break;
			}
			if(lock_spec)
			{
				fprintf(out," [lock]");  fflush(out);
				rv=USB_AtmelExecSingleCommand(
					AT_CMD_WriteLock | uint32(lock),NULL);
				if(rv) break;
			}
		} while(0);
		if(rv)
		{  fprintf(out,"Error: %s\n",USB_Status_String(rv));  }
		else
		{  fprintf(out,"\n");  }
	}
	
	return(0);
}



int AtmelDevice::DumpFuseLock(FILE * /*out*/)
{
	fprintf(stderr,
		"OOPS: AtmelDevice::DumpFuseLock() (atmel.cc) not implemented:\n"
		"      Fuse/lock reading not implemented for %s\n",name);
	return(-1);
}

	
int AtmelDevice::WriteFuseLock(FILE * /*out*/,
	char *const * /*args*/,int /*nargs*/)
{
	fprintf(stderr,
		"OOPS: AtmelDevice::WriteFuseLock() (atmel.cc) not implemented:\n"
		"      Fuse/lock writing not implemented for %s\n",name);
	return(-1);
}


void AtmelDevice::_DumpLock(FILE * /*out*/,uint8 /*lock*/)
{
	fprintf(stderr,"OOPS: Non-overridden virtual function called (_DumpLock)\n");
}
void AtmelDevice::_DumpFuseL(FILE * /*out*/,uint8 /*fuse*/)
{
	fprintf(stderr,"OOPS: Non-overridden virtual function called (_DumpFuseL)\n");
}
void AtmelDevice::_DumpFuseH(FILE * /*out*/,uint8 /*fuseH*/)
{
	fprintf(stderr,"OOPS: Non-overridden virtual function called (_DumpFuseH)\n");
}
void AtmelDevice::_DumpFuseX(FILE * /*out*/,uint8 /*fuseH*/)
{
	fprintf(stderr,"OOPS: Non-overridden virtual function called (_DumpFuseX)\n");
}

// For _DumpCalib(), there is a default implementation. 
void AtmelDevice::_DumpCalib(FILE *out,int ncalib,uint8 calib[])
{
	fprintf(out,"Calibration bytes:\n");
	for(int i=0; i<ncalib; i++)
	{
		fprintf(out,"  Calibration %d: 0x%02x   %3u\n",
			i,
			(unsigned int)calib[i],
			(unsigned int)calib[i]);
	}
}


AtmelDevice::AtmelDevice(const char *_vendor,const char *_name)
{
	vendor=_vendor;
	name=_name;
	
	flash_size=0;
	eeprom_size=0;
	
	page_size=0;
	fuse_lock_bytes=0;
	has_erase_busy=0;
	calibration_bytes=0;
}
