/*
 * bitfield.cc -- USB Atmel programmer: Bitfield with arbitrary size. 
 * 
 * Copyright (c) 2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "bitfield.h"
#include <stdio.h>
#include <ctype.h>


void XBitfield::_AllocFail()
{
	fprintf(stderr,"malloc() failed\n");
	exit(1);
}


char *XBitfield::scan16(const char *buf)
{
	_free();
	
	while(isspace(*buf)) ++buf;
	const char *start=buf;
	while(isxdigit(*buf)) ++buf;
	const char *end=buf;
	
	nbits=(end-start)*4;
	if(!nbits)  return((char*)end);
	bits=_alloc(nbits);
	for(int i=0,iend=_B2W(nbits); i<iend; i++)
	{  bits[i]=0;  }
	
	// Storage: 
	//  input: 02010510
	//  word bits[0]: 0000000002010510
	// So, 10 is the lest significant byte. 
	//  input aaaaaaaaaaaaabcccccccccccccccd
	//  word bits[0]: cccccccccccccccd
	//  word bits[1]: 00aaaaaaaaaaaaab
	// So, LSB is the first array index. The last index may not be 
	// fully populated. 
	int nn=0;  // Number of nibbles. 
	for(const char *c=end-1; c>=start; c--)
	{
		uint64 val;
		if(*c>='0' && *c<='9')
		{  val=*c-'0';  }
		else if(*c>='a' && *c<='f')
		{  val=*c-'a'+10;  }
		else if(*c>='A' && *c<='F')
		{  val=*c-'A'+10;  }
		else assert(0);
		
		assert(val>=0 && val<16);
		
		bits[nn/16]|=val<<(4*(nn%16));
		++nn;
	}
	
	return((char*)end);
}


void XBitfield::ShiftRight()
{
	if(!nbits) return;
	
	for(int i=0,iend=_B2W(nbits)-1; ; i++)
	{
		bits[i]>>=1;
		if(i>=iend)  break;
		if(bits[i+1]&1U)
		{  bits[i]|=(1LLU<<63);  }
	}
}


static const char cv_xdig[16]={
	'0','1','2','3','4','5','6','7',
	'8','9','a','b','c','d','e','f'};

void XBitfield::print(FILE *f)
{
	int nn=(nbits+3)/4;  // Number of nibbles. 
	char tmp[nn+1];
	
	char *end=tmp+nn;
	*end='\0';
	uint64 btmp;
	nn=0;
	for(char *c=end-1; c>=tmp; c--)
	{
		unsigned int val=(bits[nn/16]>>(4*(nn%16)))&0xfU;
		*c=cv_xdig[val];
		++nn;
	}
	
	fprintf(f,"%s",tmp);
}


#if 0
void TESTBitfield()
{
	const char *tests[]={
		"","1","01","0000001","1234567",
		"123456789012345","1234567890123456","12345678901234567",
		"abcdef321546789",
		"111111111111111111111111111111111111111111111",
		"1414141414141414141414141414141414141414141414",
		"198746598298374987598264987394827356928347096802938497968729384798273948756928374928374982759283749823748279596982374928734987",
		NULL };
	XBitfield bf;
	for(int i=0; tests[i]; i++)
	{
		printf("%s\n",tests[i]);
		bf.scan16(tests[i]);
		bf.ShiftRight();
		bf.print(stdout);
		printf("\n");
	}
}
#endif
