/*
 * bitfield.h -- USB Atmel programmer: Bitfield with arbitrary size. 
 * 
 * Copyright (c) 2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef _USB_ATMEL_PRG_BITFIELD_H_
#define _USB_ATMEL_PRG_BITFIELD_H_ 1

#include "config.h"
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

typedef uint64_t uint64;


// Arbitrary length bitfield. 
// Not especially performant but that's not needed since IO speed limits 
// things anyway. 
class XBitfield
{
	private:
		int nbits;
		uint64 *bits;
		
		// Convert number of bits to number of words. 
		// 0 -> 0; 1..64 -> 1
		inline int _B2W(int nb)
			{  return((nb+8*sizeof(uint64)-1)/(8*sizeof(uint64)));  }
		
		void _AllocFail();
		
		inline uint64* _alloc(int nbits)
		{
			if(!nbits) return(NULL);
			uint64 *ptr=(uint64*)malloc(_B2W(nbits)*sizeof(uint64));
			if(!ptr) _AllocFail();
			return(ptr);
		}
		inline uint64 *_free(uint64 *ptr)
			{  if(ptr) free(ptr);  return(NULL);  }
		
		inline void _free()
			{  bits=_free(bits);  nbits=0;  }
		inline void _copy(const XBitfield &b)
		{
			bits=_alloc(nbits=b.nbits);
			memcpy(bits,b.bits,_B2W(nbits)*sizeof(uint64));
		}
	public:
		// Default constructor. No content. 
		inline XBitfield() : nbits(0),bits(NULL)  {}
		// Copy constructor. 
		inline XBitfield(const XBitfield &b)
			{  _copy(b);  }
		inline ~XBitfield()
			{  _free();  }
		
		inline XBitfield &operator=(const XBitfield &b)
		{
			if(_B2W(nbits)!=_B2W(b.nbits))
			{  _free(bits);  bits=_alloc(b.nbits);  }
			memcpy(bits,b.bits,_B2W(nbits=b.nbits)*sizeof(uint64));
			return(*this);
		}
		
		// Query if last bit is set. 
		inline bool LSBIsSet() const
			{  return(nbits ? (bits[0]&1U) : 0);  }
		
		// Shift the complete value to the right by 1 bit (>>=1). 
		void ShiftRight();
		
		// Scan hex digit string in the sense of strtoul(). 
		// Will skip any amount of whitespace, then eat as many digits as 
		// it can storing them in the bitfield (content replaces current 
		// content) and then returns pointer to next char in input. 
		char *scan16(const char *buf);
		
		// Print bitfield to FILE in hex notation. 
		void print(FILE *f);
};

#endif  /* _USB_ATMEL_PRG_BITFIELD_H_ */
