/*
 * jtag.cc -- USB Atmel programmer: JTAG/SVF support. 
 * 
 * Copyright (c) 2006--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "config.h"
#include "jtag.h"
#include "bitfield.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <errno.h>

#include <assert.h>


// This is from the main program: 
extern volatile int caught_sigint;

// This is a hack from the main program. 
extern int jtag_delay_prolong_fact;


// Returns static data. 
static char *binary(uint32 x)
{
   static char rv[8*sizeof(uint32)+1];
   for(int i=0; i<(int)(8*sizeof(uint32)); i++)
   {  rv[8*sizeof(uint32)-1-i] = ((x>>i) & 1U) ? '1' : '0';  }
   rv[8*sizeof(uint32)]='\0';
   return(rv);
}


// This is a limited ID code reservoir of those devices I use or may use 
// privately. Devices are added when needed. 
// No aim for completeness in any way. 
static const char *JTAG_IDCode2Device(uint16 manufact,uint16 partnum,
	const char **part_str)
{
	*part_str=NULL;
	
	switch(manufact)
	{
		case 0x009U:
			return("Intel");
		case 0x01fU:
			switch(partnum)
			{
				case 0x5B07U: *part_str="AT91SAM7S32";  break;
				case 0x5B12U: *part_str="AT91SAM7S321"; break;
				case 0x5B06U: *part_str="AT91SAM7S64";  break;
				case 0x5B0AU: *part_str="AT91SAM7S128"; break;
				case 0x5B09U: *part_str="AT91SAM7S256"; break;
				case 0x5B1AU: *part_str="AT91SAM7S512"; break;
				//default: Do nothing. 
			}
			return("Atmel");
		case 0x049:
			switch(partnum)
			{
				case 0x9604U: *part_str="XC9572XL"; break;
				case 0x9608U: *part_str="XC95144XL"; break;
				//default: Do nothing. 
			}
			return("Xilinx");
		//default: Do nothing. 
	}
	return("unknown");
}


static void JTAG_DumpIDCode(uint32 idcode)
{
	fprintf(stderr,"JTAG ID code: 0x%08x = %s\n",idcode,binary(idcode));
	uint16 manufact=(idcode>>1)&0x7ffU;   // 11 bits
	uint16 partnum=(idcode>>12)&0xffffU;  // 16 bits
	const char *part_str=NULL;
	const char *manuf_str=JTAG_IDCode2Device(manufact,partnum,&part_str);
	fprintf(stderr,
		"  Version:   %u\n"
		"  Manufact:  0x%03x   (%s)\n"
		"  Part num:  0x%04x%s%s%s\n"
		"  Marker:    %u       (should be 1)\n"
		,(idcode>>28)  //  4 bits
		,manufact,manuf_str
		,partnum,
			part_str ? "  (" : "",part_str ? part_str : "",part_str ? ")" : ""
		,(idcode & 1U));      //  1 bit
#if 0
	// ARM TAP ID codes: http://www.arm.com/support/faqip/3843.html
	if(!(partnum & (1U<<15)))
	{
		// Seems to be an ARM core ID. 
		fprintf(stderr,
			"  ARM Core: Capability: %u%u%u\n"
			"            Family:     %u\n"
			"            Device:     0x%02x\n"
			,(partnum>>14)&1U,(partnum>>13)&1U,(partnum>>12)&1U
			,(partnum>>8)&0xfU
			,(partnum&0xffU));
	}
#endif
}


int JTagTAP::_QueryIDCode()
{
	int bit_idx[32];
	TapInit();
	// Now in RUN_TEST_IDLE. 
	
	// Read ID code as of www.fpga4fun.com/JTAG2.html: 
	// Unlike for the BYPASS instruction, the IR value for the IDCODE is not 
	// standard. Fortunately, each time the TAP controller goes into 
	// Test-Logic-Reset, it goes into IDCODE mode (loads the IDCODE into DR).
// FIXME: Create own function for this! (DR scan, IR scan)
	clock(TMS);  // Sel DR scan
	clock(0);    // Capture DR
	clock(0);    // Shift DR
	
	// Read ID code. 
	uint32 idcode=0U;
	for(int jj=0; jj<31; jj++)
	{  clock(0,&idcode,jj);  }  // bit jj
	clock(TMS,&idcode,31);  // bit 31; Exit DR
	clock(TMS);  // Update DR
	clock(0);  // Back in RTI. 
	
	TapReset();
	// Back in TEST_LOGIC_RESET mode. 
	
	// USB communication...
	int rv=wfb.flush(usbatmelprg_jtag_speed);
	if(rv)
	{  fprintf(stderr,"Failed to read JTAG ID code.\n");  return(-1);  }
	else
	{  JTAG_DumpIDCode(idcode);  }
	
	return(0);
}


int JTagTAP::_ScanIDCodes()
{
	const int nbits=1000;
	
	fprintf(stderr,"Detecting ... ");
	
	// See http://www.fpga4fun.com/JTAG3.html for more information on 
	// how this is supposed to work. 
	TapInit();
	// Go to SHIFT_IR: 
	//TapMoveTo(SHIFT_IR);
	clock(TMS);
	clock(TMS);
	clock(0);
	clock(0);
	
	// Send plenty of ones into the IR registers. 
	// That makes sure all devices are in BYPASS!
	for(int i=1; i<nbits; i++)
	{  clock(TDI);  }
	clock(TMS|TDI);  // Last bit needs to have TMS active, to exit shift-IR. 
	
	// We are in Exit1-IR, go to Shift-DR. 
	clock(TMS);
	clock(TMS);
	clock(0);
	clock(0);
	
	// Send plenty of zeros into the DR registers to flush them. 
	for(int i=0; i<nbits; i++)
	{  clock(0);  }
	
	// Now send ones until we receive one back. 
	uint8 resp_bits[nbits];
	memset(resp_bits,0,sizeof(nbits));
	for(int i=0; i<nbits; i++)
	{  clock(TDI,&resp_bits[i],0);  }
	
	// USB communication...
	int rv=wfb.flush(usbatmelprg_jtag_speed);
	if(rv)
	{  fprintf(stderr,"Failed to scan JTAG chain.\n");  return(-1);  }
	
	int n_devices=-1;
	for(int i=0; i<nbits; i++)
	{
		if(resp_bits[i])
		{  n_devices=i;  break;  }
	}
	fprintf(stderr,"detected %d device%s.\n",
		n_devices,n_devices==1 ? "" : "s");
	
	if(n_devices)
	{
		// Read all the ID codes: 
		TapInit();
		// Go to SHIFT_DR. 
		clock(TMS);
		clock(0);
		clock(0);
	}
	// Read the ID codes. 
	for(int i=0; i<n_devices; i++)
	{
		uint32 idcode=0U;
		for(int jj=0; jj<32; jj++)
		{  clock(0,&idcode,jj);  }  // bit jj
		
		int rv=wfb.flush(usbatmelprg_jtag_speed);
		if(rv)
		{  fprintf(stderr,"Failed to scan JTAG chain.\n");  return(-1);  }
		
		fprintf(stderr,"Device[%d]: ",i);
		JTAG_DumpIDCode(idcode);
	}
	
	return(0);
}


//==============================================================================

template<typename UI>void JTagTAP::JTagIOQueued(
	UI val,int nbits,bool is_instruction,
	UI expect_mask,UI expect_val)
{
	assert(nbits<=(int)(8*sizeof(UI)));
	
	// Expecting to be in RUN_TEST_IDLE state. 
	clock(TMS);         // SELECT_DR_SCAN. 
	if(is_instruction)
	{  clock(TMS);  }   // SELECT_IR_SCAN. 
	
	clock(0);
	clock(0);
	// Went via CAPTURE_DR/CAPTURE_IR to SHIFT_DR/SHIFT_IR. 
	
	for(int bc=0; bc<nbits; bc++)
	{
		uint8 state=0;
		
		// If this is the last bit, set TMS high to leave the shift state 
		// at the next clock just at the same time we clock in the last bit. 
		if(bc+1==nbits)
		{  state|=TMS;  }
		
		if(val & 0x01U)
		{  state|=TDI;  }
		
		if(expect_mask & 0x01U)
		{  clock(state,expect_val & 0x01U);  }
		else
		{  clock(state);  }
		
		val>>=1;
		expect_mask>>=1;
		expect_val>>=1;
	}
	
	clock(TMS);
	// We're now in the update state. 
	
	clock(0);
	// Back in RUN_TEST_IDLE state. 
}

void JTagTAP::JTagIOQueuedBF(
	const XBitfield *_val,int nbits,bool is_instruction,
	const XBitfield *_expect_mask,const XBitfield *_expect_val)
{
	// Expecting to be in RUN_TEST_IDLE state. 
	clock(TMS);         // SELECT_DR_SCAN. 
	if(is_instruction)
	{  clock(TMS);  }   // SELECT_IR_SCAN. 
	
	clock(0);
	clock(0);
	// Went via CAPTURE_DR/CAPTURE_IR to SHIFT_DR/SHIFT_IR. 
	
	// Copy values so that we may modify them. 
	XBitfield val(*_val);
	XBitfield expect_mask;  if(_expect_mask) expect_mask=*_expect_mask;
	XBitfield expect_val;   if(_expect_val)  expect_val= *_expect_val;
	
	for(int bc=0; bc<nbits; bc++)
	{
		uint8 state=0;
		
		// If this is the last bit, set TMS high to leave the shift state 
		// at the next clock just at the same time we clock in the last bit. 
		if(bc+1==nbits)
		{  state|=TMS;  }
		
		if(val.LSBIsSet())
		{  state|=TDI;  }
		
		if(_expect_mask && expect_mask.LSBIsSet())
		{  clock(state,_expect_val ? expect_val.LSBIsSet() : 0);  }
		else
		{  clock(state);  }
		
		val.ShiftRight();
		if(_expect_mask) expect_mask.ShiftRight();
		if(_expect_val)  expect_val.ShiftRight();
	}
	
	clock(TMS);
	// We're now in the update state. 
	
	clock(0);
	// Back in RUN_TEST_IDLE state. 
}

//==============================================================================

int JTagTAP::_SVFProgramFile(const char *filename)
{
	int errors=0;
	int io_errors=0;
	
	FILE *fp=fopen(filename,"r");
	if(!fp)
	{  fprintf(stderr,"Opening \"%s\": %s\n",filename,strerror(errno));
		return(-1);  }
	
	assert(wfb.IsEmpty());
	wfb.nblocks_sent=0;
	wfb.totsize_sent=0;
	
	TapInit();
	
	timeval tv_start;
	gettimeofday(&tv_start,NULL);
	
	// FIXME: 
	// - Max line length for SVF files is 256 but maybe it allows 
	//   to continue on the next line until the ";"... 
	// I'll implement things once they are needed. 
	char linebuf[256+1];
	XBitfield v_tdi,v_tdo,v_mask,v_smask;
	int lineno=0;
	long tot_usec_slept=0;
	while(!feof(fp))
	{
		*linebuf=0;
		if(!fgets(linebuf,sizeof(linebuf),fp))
		{
			if(feof(fp))  break;
			fprintf(stderr,"Reading \"%s\": %s\n",filename,
				strerror(ferror(fp)));
			++errors;
			break;
		}
		++lineno;
		char *lb=linebuf;
		if(*lb=='\r')  ++lb;
		if(!*lb)  continue;
		// Skip comments. 
		if(lb[0]=='/' && lb[1]=='/')
		{
			if(verbose>1)
			{  printf("%05d: %s",lineno,lb);  }
			continue;
		}
		// Interprete commands. 
		// Several of them are not fully supported. 
		if(!strncmp(lb,"TRST OFF",8))
		{
			// The programmer doesn't support the TRST pin anyways 
			// so switching it off is allright. 
			continue;
		}
		if(!strncmp(lb,"ENDIR IDLE",10) || !strncmp(lb,"ENDDR IDLE",10))
		{
			// End IR and DR scans with IDLE state (?). 
			// This is what we do anyway. 
			// Don't know what these mean exactly but it seems 
			// we can simply ignore them. 
			continue;
		}
		if(!strncmp(lb,"FREQUENCY ",10))
		{
			// We don't care about frequency settings. 
			// We're running at 1MHz by default and no device should ever 
			// require less speed. After all, the user can change this. 
			continue;
		}
		if(!strncmp(lb,"TIR ",4) || 
		   !strncmp(lb,"HIR ",4) ||
		   !strncmp(lb,"TDR ",4) ||
		   !strncmp(lb,"HDR ",4) )
		{
			// So, this is actually important. 
			// We don't support chains yet so these must all specify 0. 
			lb+=4;
			int val=strtol(lb,&lb,10);
			if(val!=0)
			{
				fprintf(stderr,"%d: %s <-- Not supported; giving up.\n",
					lineno,lb-4);
				++errors;
				break;
			}
			continue;
		}
		if(!strncmp(lb,"STATE ",6))
		{
			if(!strncmp(lb,"STATE RESET IDLE",16) || 
			   !strncmp(lb,"STATE IDLE",10))
			{
				if(verbose>1)  printf("%05d: %s",lineno,lb);
				TapInit();
				if(wfb.flush(usbatmelprg_jtag_speed,'s')<0)
				{  ++io_errors;  break;  }
			}
			else if(!strncmp(lb,"STATE RESET",11))
			{
				if(verbose>1)  printf("%05d: %s",lineno,lb);
				TapReset();
				if(wfb.flush(usbatmelprg_jtag_speed,'s')<0)
				{  ++io_errors;  break;  }
			}
			else
			{
				fprintf(stderr,"%d: %s <-- Not supported; giving up.\n",
					lineno,lb);
				++errors;
				break;
			}
		}
		else if(!strncmp(lb,"SIR ",4) || !strncmp(lb,"SDR ",4))
		{
			bool is_instruction;
			switch(lb[1])
			{
				case 'I':  is_instruction=1;  break;
				case 'D':  is_instruction=0;  break;
				default:  assert(0);
			}
			
			if(verbose>1)
			{  printf("%05d: IN: %s",lineno,lb);  }
			lb+=4;
			int nbits=strtol(lb,&lb,10);
			assert(lb);
			bool has_tdi=0;
			bool has_tdo=0;
			bool has_mask=0;
			bool has_smask=0;
			for(;;)
			{
				if(*lb==')')  ++lb;
				while(isspace(*lb))  ++lb;
				if(!*lb)  break;
				if(*lb==';')  break;
				if(!strncmp(lb,"TDI",3))  // "TDI ("
				{  lb+=5;  lb=v_tdi.scan16(lb);  has_tdi=1;  }
				else if(!strncmp(lb,"TDO",3))
				{  lb+=5;  lb=v_tdo.scan16(lb);  has_tdo=1;  }
				else if(!strncmp(lb,"MASK",4))
				{  lb+=6;  lb=v_mask.scan16(lb);  has_mask=1;  }
				else if(!strncmp(lb,"SMASK",5))
				{  lb+=7;  lb=v_smask.scan16(lb);  has_smask=1;  }
				else
				{  fprintf(stderr,"Parse error: <%s>\n",lb);  assert(0);  }
			}
			if(verbose>1)
			{
				printf("       PA: S%cR %d ",is_instruction ? 'I' : 'D',nbits);
#if 0
				if(has_tdi)   { printf("TDI (%0*Lx) ",(nbits+7)/8*2,v_tdi);  }
				if(has_tdo)   { printf("TDO (%0*Lx) ",(nbits+7)/8*2,v_tdo);  }
				if(has_mask)  { printf("MASK (%0*Lx) ",(nbits+7)/8*2,v_mask);  }
				if(has_smask) { printf("SMASK (%0*Lx) ",(nbits+7)/8*2,v_smask); }
#else
				if(has_tdi)   { printf("TDI (");   v_tdi.print(stdout);   printf(") "); }
				if(has_tdo)   { printf("TDO (");   v_tdo.print(stdout);   printf(") "); }
				if(has_mask)  { printf("MASK (");  v_mask.print(stdout);  printf(") "); }
				if(has_smask) { printf("SMASK ("); v_smask.print(stdout); printf(") "); }
#endif
				printf(";\n");
			}
			
			JTagIOQueuedBF(&v_tdi,nbits,is_instruction,
				/*expect_mask=*/has_tdo ? &v_mask : NULL,
				/*expect_val=*/ has_tdo ? &v_tdo  : NULL);
			
			// Flush if we have enough blocks pending. 
			// (The last block may not be completely full but oh well...)
			if(wfb.NPending()>=48)
			{
				if(wfb.flush(usbatmelprg_jtag_speed,'s')<0)
				{  ++io_errors;  break;  }
			}
		}
		else if(!strncmp(lb,"RUNTEST ",8))
		{
			// If the delay is short, we simply make the apropriate number 
			// of clock cycles. Otherwise, we wait. 
			long usec=strtol(lb+8,NULL,10);
			if(usec>16)
			{
				// FIXME: There is no point in doing delays which are shorter 
				//        than our USB transmission frame rate. 
				if(verbose>1) printf("%05d: RUNTEST %ld TCK\n",lineno,usec);
				if(wfb.flush(usbatmelprg_jtag_speed,'s')<0)  // Important, otherwise it's pointless!
				{  ++io_errors;  break;  }
				printf("[delay %ld usec]",usec); fflush(stdout);
				
				errno=0;
				// Interestingly, I get errors when using usleep(usec) 
				// alone. Must be somewhat longer. 
				//   usec+usec/16 gives several errors and one at flash erase. 
				//   usec+usec/8  gives 1 error at flash erase. 
				//   usec+usec/4  gives no errors. 
				long real_delay=usec+jtag_delay_prolong_fact*usec/10;
				usleep(real_delay);
				tot_usec_slept+=real_delay;
			}
			else
			{
				for(int i=usec; i; i--)
				{  clock(0);  }
			}
		}
		else
		{
			fprintf(stderr,"%05d: Parse error: unknown <%s>\n",lineno,lb);
			++errors;
			break;
		}
		
		if(caught_sigint)  break;
	}
	
	fclose(fp);
	
	if(io_errors)
	{  /*printf("IO ERRORS\n");*/  }
	else
	{
		TapReset();
		if(wfb.flush(usbatmelprg_jtag_speed,'s')<0)
		{  ++io_errors;  }
		printf("\n");
	}
	
	if(!io_errors)
	{
		timeval tv_end;
		gettimeofday(&tv_end,NULL);
		long msec=(tv_end.tv_sec-tv_start.tv_sec)*1000+
	        	  ((int)tv_end.tv_usec-(int)tv_start.tv_usec+500)/1000;
		long worked_msec = msec-(tot_usec_slept+500)/1000;
		fprintf(stderr,"Elapsed: %ld msec (slept %ld msec, worked %ld msec); "
			"rate: %f msec/pack\n",
			msec,(tot_usec_slept+500)/1000,worked_msec,
			(double)worked_msec/(double)wfb.nblocks_sent);
		// Compute theoretical minimum time: All the sleep times accumulated 
		// and assume 1usec per transmitted bit which is the (max allowed) 1MHz 
		// JTAG port clock frequency. 
		//fprintf(stderr,"Transmitted %d bits; min. time required: %ld msec\n",
		//	wfb.totsize_sent,
		//	(tot_usec_slept+wfb.totsize_sent+500)/1000);
	}
	
	return(errors+io_errors);
}

//==============================================================================

JTagTAP::~JTagTAP()
{
	assert(wfb.IsEmpty());
}
