/*
 * jtag.h -- USB Atmel programmer: JTAG/SVF support. 
 * 
 * Copyright (c) 2006--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef _USB_ATMEL_PRG_JTAG_H_
#define _USB_ATMEL_PRG_JTAG_H_ 1

#include "../cable/usbio.h"
#include "../cable/waveformbuf.h"
#include <assert.h>
#include <string.h>


// JTAG speed value. 
extern uint8 usbatmelprg_jtag_speed;

// bitfield.h: 
class XBitfield;


// TAP = test access port
// JTAG = joint test action group
// 
// .bit files (bitgen) are only available for FPGAs and not for CPLDs. 
// 
// The JTAG Clock for the XC9500/XL devices, TCK, has a maximum frequency 
// of 10 MHz.
// 
// The 1149.1 standard defines that TDI is valid and shifted in (and TDO valid
// and shifted out) only in the Shift-DR or Shift-IR states. The Shift-IR state
// selects the device instruction register between TDI and TDO. Depending on the
// instruction selected different data registers are activated. When in
// Shift-DR, the data register appropriate for the previously entered
// instruction is selected between TDI and TDO. The default data register is the
// mandatory 1-bit bypass register.
// 
// The JTAG TAP Controller is a 16-state finite state machine, that controls the
// scanning of data into the various registers of the JTAG architecture. The
// state of the TMS pin at the rising edge of TCK is responsible for determining
// the sequence of state transitions. There are two state transition paths for
// scanning the signal at TDI into the device, one for shifting in an
// instruction to the instruction register and one for shifting data into the
// active data register as determined by the current instruction.
// 
// TDO changes state on the falling edge of TCK and is only active during the 
// shifting of instructions or data through the device. This pin is 3-stated 
// at all other times.
// 
// When using Xilinx XC9500/XL/XV parts, the TDO values predicted reflect the
// status of the just completed operation (which could be an erase or a program
// operation). If the status is not the success status (which is the value
// predicted as the TDO value in the generated SVF file), then the following
// 1149.1 TAP controller state transition sequence must be followed to correctly
// erase and program the XC9500/XL/XV device (assuming the TDO validation failure
// is detected in the EXIT1-DR state):
//   1. EXIT1-DR
//   2. PAUSE-DR
//   3. EXIT2-DR
//   4. SHIFT-DR
//   5. EXIT1-DR
//   6. UPDATE-DR
//   7. RUN-TEST/IDLE
class JTagTAP
{
	public:
		// Public waveform buffer used by the JTAG routines to store the 
		// TCK/TMS/TDO values to be emitted for later transmission over the 
		// USB. 
		WaveFormBuffer wfb;
		
		// Bit values for convenience. 
		static const uint8 TDI=WaveFormBuffer::WB_TDI;
		static const uint8 TMS=WaveFormBuffer::WB_TMS;
		
		// Functions for convenience. 
		inline void clock(uint8 b)              { wfb.clock(b);        }
		inline void clock(uint8 b,uint8 expect) { wfb.clock(b,expect); }
		inline void clock(uint8 bits,uint8  *dest,uint8 bidx)  { wfb.clock(bits,dest,bidx); }
		inline void clock(uint8 bits,uint16 *dest,uint8 bidx)  { wfb.clock(bits,dest,bidx); }
		inline void clock(uint8 bits,uint32 *dest,uint8 bidx)  { wfb.clock(bits,dest,bidx); }
		inline void clock(uint8 bits,uint64 *dest,uint8 bidx)  { wfb.clock(bits,dest,bidx); }
		
		enum JtagState
		{
			// Test-Logic-Reset.  This state is entered on power-up of the
			// device whenever at least five clocks of TCK occur with TMS held
			// high. Entry into this state resets all JTAG logic to a state
			// such that it will not interfere with the normal component logic,
			// and causes the IDCODE instruction to be forced into the
			// instruction register.
			TEST_LOGIC_RESET=0,
			
			// Run-Test-Idle.  This state allows certain operations to occur
			// depending on the current instruction. For the XC9500/XL/XV
			// family, this state causes generation of the program, verify and
			// erase pulses when the associated in-system programming (ISP)
			// instruction is active.
			RUN_TEST_IDLE,
			
			// Select-DR-Scan.  This is a temporary state entered prior to
			// performing a scan operation on a data register or in passing to
			// the Select-IR-Scan state.
			SELECT_DR_SCAN,
			
			// Select-IR-Scan.  This is a temporary state entered prior to
			// performing a scan operation on the instruction register or in
			// returning to the Test-Logic-Reset state.
			SELECT_IR_SCAN,
			
			// Capture-DR.  This state allows data to be loaded from parallel
			// inputs into the data register selected by the current
			// instruction on the rising edge of TCK. If the selected data
			// register does not have parallel inputs, the register retains its
			// state.
			CAPTURE_DR,
			
			// Shift-DR.  This state shifts the data, in the currently selected
			// register, towards TDO by one stage on each rising edge of TCK
			// after entering this state.
			SHIFT_DR,
			
			// Exit1-DR.  This is a temporary state that allows the option of
			// passing on to the Pause-DR state or transitioning directly to
			// the Update-DR state.
			EXIT1_DR,
			
			// Pause-DR.  This is a wait state that allows shifting of data to
			// be temporarily halted.
			PAUSE_DR,
			
			// Exit2-DR.  This is a temporary state that allows the option of
			// passing on to the Update-DR state or returning to the Shift-DR
			// state to continue shifting in data.
			EXIT2_DR,
			
			// Update-DR.  This state causes the data contained in the
			// currently selected data register to be loaded into a latched
			// parallel output (for registers that have such a latch) on the
			// falling edge of TCK after entering this state. The parallel
			// latch prevents changes at the parallel output of these registers
			// from occurring during the shifting process.
			UPDATE_DR,
			
			// Capture-IR.  This state allows data to be loaded from parallel
			// inputs into the instruction register on the rising edge of TCK.
			// The least two significant bits of the parallel inputs must have
			// the value 01 as defined by IEEE Std. 1149.1, and the remaining 6
			// bits are either hard-coded or used for monitoring of the
			// security and data protect bits.
			CAPTURE_IR,
			
			// Shift-IR.  This state shifts the values in the instruction
			// register towards TDO by one stage on each rising edge of TCK
			// after entering this state.
			SHIFT_IR,
			
			// Exit1-IR.  This is a temporary state that allows the option of
			// passing on to the Pause-IR state or transitioning directly to
			// the Update-IR state.
			EXIT1_IR,
			
			// Pause-IR.  This is a wait state that allows shifting of the
			// instruction to be temporarily halted.
			PAUSE_IR,
			
			// Exit2-IR.  This is a temporary state that allows the option of
			// passing on to the Update-IR state or returning to the Shift-IR
			// state to continue shifting in data.
			EXIT2_IR,
			
			// Update-IR.  This state causes the values contained in the
			// instruction register to be loaded into a latched parallel output
			// on the falling edge of TCK after entering this state. The
			// parallel latch prevents changes at the parallel output of the
			// instruction register from occurring during the shifting process.
			UPDATE_IR,
		};
		
	private:
		
		// Current TAP state. 
		JtagState curr_state;
		
		// Internally used by corresponding static functions. 
		int _QueryIDCode();
		int _ScanIDCodes();
		int _SVFProgramFile(const char *filename);
		
	public:
		JTagTAP() : wfb()
			{  curr_state=TEST_LOGIC_RESET;  }
		~JTagTAP();
		
		// Bring the TAP controller into (TEST_LOGIC_) RESET state. 
		void TapReset()
		{
			// Move into TEST_LOGIC_RESET state from any other possible state. 
			for(int i=0; i<5; i++)
			{  clock(TMS);  }
			curr_state=TEST_LOGIC_RESET;
		}
		// Init for the TAP controller to bring it from any unknown state 
		// into the RUN-TEST/IDLE state. 
		void TapInit()
		{
			TapReset();   // Now in TEST_LOGIC_RESET. 
			clock(0); // Now in RUN_TEST_IDLE. 
			curr_state=RUN_TEST_IDLE;
		}
		
		// Make the necessary actions to bring the TAP controller from 
		// the current state into the passed state "to". 
		void TapMoveTo(JtagState to);
		
		// Queue JTAG IO with possible spec of expected readback value. 
		template<typename UI>void JTagIOQueued(
			UI data,int nbits,bool is_instruction,
			UI expect_mask=(UI)0,UI expect_val=(UI)0);
		void JTagIOQueuedBF(
			const XBitfield *data,int nbits,bool is_instruction,
			const XBitfield *expect_mask=NULL,
			const XBitfield *expect_val=NULL);
		
		// Read ID code. Should work for any device as long as it is only one. 
		static int QueryIDCode()
			{  JTagTAP jtag;  return(jtag._QueryIDCode());  }
		// Attempt to scan ID codes of all devices in the JTAG chain. 
		static int ScanIDCodes()
			{  JTagTAP jtag;  return(jtag._ScanIDCodes());  }
		// Download/program using SVF file. 
		// Returns error status. 
		static int SVFProgramFile(const char *filename)
			{  JTagTAP jtag;  return(jtag._SVFProgramFile(filename));  }
};

#endif  /* _USB_ATMEL_PRG_JTAG_H_ */
